from functools import wraps
from typing import Callable, Optional, Union, Type

from lxml import etree

from ... import ast


Element = etree._Element
ElementTree = etree._ElementTree
ParseInput = Union[etree._Element, etree._ElementTree, str]


class NodeParsingError(ValueError):
    pass


class Missing:
    pass


def handle(*tags: str, namespace: Union[str, Type[Missing]] = Missing,
           subiter: bool = True) -> Callable:
    """ Function-decorator to mark a class function as a handler for a
        given node type.
    """
    assert tags

    @wraps(handle)
    def inner(func):
        func.handles_tags = tags
        func.namespace = namespace
        func.subiter = subiter
        return func

    return inner


def handle_namespace(namespace: str, subiter: bool = True) -> Callable:
    """ Function-decorator to mark a class function as a handler for a
        given namespace.
    """

    @wraps(handle)
    def inner(func):
        func.handles_namespace = namespace
        func.subiter = subiter
        return func

    return inner


class XMLParserMeta(type):
    def __init__(cls, name, bases, dct):
        cls_values = [(cls, dct.values())]
        cls_namespace = getattr(cls, 'namespace', None)

        for base in bases:
            cls_namespace = cls_namespace or getattr(base, 'namespace', None)
            cls_values.append((base, base.__dict__.values()))

        tag_map = {}
        namespace_map = {}
        for cls_, values in cls_values:
            for value in values:
                if hasattr(value, 'handles_tags'):
                    for handled_tag in value.handles_tags:
                        namespace = value.namespace
                        if namespace is Missing:
                            namespace = getattr(cls_, 'namespace', None) \
                                or cls_namespace
                        if namespace:
                            if isinstance(namespace, (list, tuple)):
                                namespaces = namespace
                            else:
                                namespaces = [namespace]

                            for namespace in namespaces:
                                full_tag = f'{{{namespace}}}{handled_tag}'
                                tag_map[full_tag] = value
                        else:
                            tag_map[handled_tag] = value

                if hasattr(value, 'handles_namespace'):
                    namespace_map[value.handles_namespace] = value

        cls.tag_map = tag_map
        cls.namespace_map = namespace_map


class XMLParser(metaclass=XMLParserMeta):
    namespace: Optional[str] = None
    tag_map: dict
    namespace_map: dict

    def parse(self, input_: ParseInput) -> ast.Node:
        if isinstance(input_, Element):
            root = input_
        elif isinstance(input_, ElementTree):
            root = input_.getroot()
        else:
            root = etree.fromstring(input_)

        return self._evaluate_node(root)

    def _evaluate_node(self, node: etree._Element) -> ast.Node:
        qname = etree.QName(node.tag)
        if node.tag in self.tag_map:
            parse_func = self.tag_map[node.tag]
        elif qname.namespace in self.namespace_map:
            parse_func = self.namespace_map[qname.namespace]
        else:
            raise NodeParsingError(f"Cannot parse XML tag {node.tag}")

        if parse_func.subiter:
            sub_nodes = [
                self._evaluate_node(child)
                for child in node.iterchildren()
            ]
            return parse_func(self, node, *sub_nodes)
        else:
            return parse_func(self, node)



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2019 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------


from dataclasses import dataclass
from datetime import date, time, datetime, timedelta
from typing import Any, List, Union, Optional

from pygeoif.geometry import as_shape


@dataclass
class Geometry:
    geometry: dict

    @property
    def __geo_interface__(self):
        return self.geometry

    def __eq__(self, o: object) -> bool:
        return (
            as_shape(self).__geo_interface__ == as_shape(o).__geo_interface__
        )


@dataclass
class Envelope:
    x1: float
    x2: float
    y1: float
    y2: float

    @property
    def geometry(self):
        return {
            "type": "Polygon",
            "coordinates": [[
                [self.x1, self.y1],
                [self.x1, self.y2],
                [self.x2, self.y2],
                [self.x2, self.y1],
                [self.x1, self.y1]
            ]]
        }

    @property
    def __geo_interface__(self):
        return self.geometry

    def __eq__(self, o: object) -> bool:
        return (
            as_shape(self).__geo_interface__ == as_shape(o).__geo_interface__
        )


@dataclass
class Interval:
    start: Optional[Union[date, datetime, timedelta]] = None
    end: Optional[Union[date, datetime, timedelta]] = None

    def get_sub_nodes(self) -> List[Any]:  # TODO: find way to type this
        return [self.start, self.end]


# used for handler declaration
LITERALS = (list, str, float, int, bool, datetime, date, time, timedelta)

# used for type checking

SpatialValueType = Union[Geometry, Envelope]

TemporalValueType = Union[date, datetime, timedelta, Interval]

ValueType = Union[
    SpatialValueType,
    TemporalValueType,
    bool,
    float,
    int,
    str,
]



# Common configurations for cql2 parsers and evaluators.
from typing import Dict, Type, Union
from . import ast

# https://github.com/opengeospatial/ogcapi-features/tree/master/cql2


COMPARISON_MAP: Dict[str, Type[ast.Node]] = {
    "=": ast.Equal,
    "eq": ast.Equal,
    "<>": ast.NotEqual,
    "!=": ast.NotEqual,
    "ne": ast.NotEqual,
    "<": ast.LessThan,
    "lt": ast.LessThan,
    "<=": ast.LessEqual,
    "lte": ast.LessEqual,
    ">": ast.GreaterThan,
    "gt": ast.GreaterThan,
    ">=": ast.GreaterEqual,
    "gte": ast.GreaterEqual,
    "like": ast.Like,
}

SPATIAL_PREDICATES_MAP: Dict[str, Type[ast.SpatialComparisonPredicate]] = {
    "s_intersects": ast.GeometryIntersects,
    "s_equals": ast.GeometryEquals,
    "s_disjoint": ast.GeometryDisjoint,
    "s_touches": ast.GeometryTouches,
    "s_within": ast.GeometryWithin,
    "s_overlaps": ast.GeometryOverlaps,
    "s_crosses": ast.GeometryCrosses,
    "s_contains": ast.GeometryContains,
}

TEMPORAL_PREDICATES_MAP: Dict[str, Type[ast.TemporalPredicate]] = {
    "t_before": ast.TimeBefore,
    "t_after": ast.TimeAfter,
    "t_meets": ast.TimeMeets,
    "t_metby": ast.TimeMetBy,
    "t_overlaps": ast.TimeOverlaps,
    "t_overlappedby": ast.TimeOverlappedBy,
    "t_begins": ast.TimeBegins,
    "t_begunby": ast.TimeBegunBy,
    "t_during": ast.TimeDuring,
    "t_contains": ast.TimeContains,
    "t_ends": ast.TimeEnds,
    "t_endedby": ast.TimeEndedBy,
    "t_equals": ast.TimeEquals,
    "t_intersects": ast.TimeOverlaps,
}


ARRAY_PREDICATES_MAP: Dict[str, Type[ast.ArrayPredicate]] = {
    "a_equals": ast.ArrayEquals,
    "a_contains": ast.ArrayContains,
    "a_containedby": ast.ArrayContainedBy,
    "a_overlaps": ast.ArrayOverlaps,
}

ARITHMETIC_MAP: Dict[str, Type[ast.Arithmetic]] = {
    "+": ast.Add,
    "-": ast.Sub,
    "*": ast.Mul,
    "/": ast.Div,
}

CONDITION_MAP: Dict[str, Type[ast.Node]] = {
    "and": ast.And,
    "or": ast.Or,
    "not": ast.Not,
    "isNull": ast.IsNull,
}

BINARY_OP_PREDICATES_MAP: Dict[
    str,
    Union[
        Type[ast.Node],
        Type[ast.Comparison],
        Type[ast.SpatialComparisonPredicate],
        Type[ast.TemporalPredicate],
        Type[ast.ArrayPredicate],
        Type[ast.Arithmetic],
    ],
] = {
    **COMPARISON_MAP,
    **SPATIAL_PREDICATES_MAP,
    **TEMPORAL_PREDICATES_MAP,
    **ARRAY_PREDICATES_MAP,
    **ARITHMETIC_MAP,
    **CONDITION_MAP,
}


def get_op(node: ast.Node) -> Union[str, None]:
    # Get the cql2 operator string from a node.
    for k, v in BINARY_OP_PREDICATES_MAP.items():
        if isinstance(node, v):
            return k
    return None



from ... import ast
from .base import FESBaseParser
from .util import handle, ParseInput, Element


class FES11Parser(FESBaseParser):
    namespace = 'http://www.opengis.net/ogc'

    @handle('Add')
    def add(self, node: Element, lhs: ast.ScalarAstType,
            rhs: ast.ScalarAstType) -> ast.Node:
        return ast.Add(lhs, rhs)

    @handle('Sub')
    def sub(self, node: Element, lhs: ast.ScalarAstType,
            rhs: ast.ScalarAstType) -> ast.Node:
        return ast.Sub(lhs, rhs)

    @handle('Mul')
    def mul(self, node: Element, lhs: ast.ScalarAstType,
            rhs: ast.ScalarAstType) -> ast.Node:
        return ast.Mul(lhs, rhs)

    @handle('Div')
    def div(self, node: Element, lhs: ast.ScalarAstType,
            rhs: ast.ScalarAstType) -> ast.Node:
        return ast.Div(lhs, rhs)


def parse(input_: ParseInput) -> ast.Node:
    return FES11Parser().parse(input_)



from typing import Union

from lxml import etree

from ... import ast
from .v11 import FES11Parser
from .v20 import FES20Parser
from .util import Element, ElementTree


def parse(xml: Union[str, Element, ElementTree]) -> ast.Node:
    if isinstance(xml, str):
        root = etree.fromstring(xml)
    else:
        root = xml

    # decide upon namespace which parser to use
    namespace = etree.QName(root).namespace
    if namespace == FES11Parser.namespace:
        return FES11Parser().parse(root)
    elif namespace == FES20Parser.namespace:
        return FES20Parser().parse(root)

    raise ValueError(f'Unsupported namespace {namespace}')



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2019 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------


from operator import and_, or_, add, sub, mul, truediv
from datetime import datetime, timedelta
from functools import reduce
from typing import Dict, List, Optional, Union

from django.db.models import Q, F, Value
from django.db.models.expressions import Expression

from django.contrib.gis.gdal import SpatialReference
from django.contrib.gis.geos import Polygon
from django.contrib.gis.measure import D

ArithmeticType = Union[Expression, F, Value, int, float]

# ------------------------------------------------------------------------------
# Filters
# ------------------------------------------------------------------------------


def combine(sub_filters: List[Q], combinator: str = "AND") -> Q:
    """ Combine filters using a logical combinator
    """
    op = and_ if combinator == "AND" else or_
    return reduce(lambda acc, q: op(acc, q) if acc else q, sub_filters)


def negate(sub_filter: Q) -> Q:
    """ Negate a filter, opposing its meaning.
    """
    return ~sub_filter


OP_TO_COMP = {
    "<": "lt",
    "<=": "lte",
    ">": "gt",
    ">=": "gte",
    "<>": None,
    "=": "exact"
}

INVERT_COMP: Dict[Optional[str], str] = {
    "lt": "gt",
    "lte": "gte",
    "gt": "lt",
    "gte": "lte",
}


def compare(lhs: Union[F, Value], rhs: Union[F, Value], op: str,
            mapping_choices: Optional[Dict[str, Dict[str, str]]] = None) -> Q:
    """ Compare a filter with an expression using a comparison operation

        :param lhs: the field to compare
        :type lhs: :class:`django.db.models.F`
        :param rhs: the filter expression
        :type rhs: :class:`django.db.models.F`
        :param op: a string denoting the operation. one of ``"<"``, ``"<="``,
                   ``">"``, ``">="``, ``"<>"``, ``"="``
        :type op: str
        :param mapping_choices: a dict to lookup potential choices for a
                                certain field.
        :type mapping_choices: dict[str, str]
        :return: a comparison expression object
        :rtype: :class:`django.db.models.Q`
    """
    comp = OP_TO_COMP[op]

    # if the left hand side is not a field reference, the comparison
    # can be be inverted to try if the right hand side is a field
    # reference.
    if not isinstance(lhs, F):
        lhs, rhs = rhs, lhs
        comp = INVERT_COMP.get(comp, comp)

    # if neither lhs and rhs are fields, we have to fail here
    if not isinstance(lhs, F):
        raise ValueError(f'Unable to compare non-field {lhs}')

    field_name = lhs.name

    if mapping_choices and field_name in mapping_choices:
        try:
            if isinstance(rhs, str):
                rhs = mapping_choices[field_name][rhs]
            elif hasattr(rhs, 'value'):
                rhs = Value(mapping_choices[field_name][rhs.value])

        except KeyError as e:
            raise AssertionError("Invalid field value %s" % e)

    if comp:
        return Q(**{"%s__%s" % (lhs.name, comp): rhs})
    return ~Q(**{field_name: rhs})


def between(lhs: F, low: Union[F, Value], high: Union[F, Value],
            not_: bool = False) -> Q:
    """ Create a filter to match elements that have a value within a certain
        range.

        :param lhs: the field to compare
        :type lhs: :class:`django.db.models.F`
        :param low: the lower value of the range
        :type low:
        :param high: the upper value of the range
        :type high:
        :param not_: whether the range shall be inclusive (the default) or
                     exclusive
        :type not_: bool
        :return: a comparison expression object
        :rtype: :class:`django.db.models.Q`
    """
    q = Q(**{"%s__range" % lhs.name: (low, high)})
    return ~q if not_ else q


def like(lhs: F, pattern: str, nocase: bool = False, not_: bool = False,
         mapping_choices: Optional[Dict[str, Dict[str, str]]] = None) -> Q:
    """ Create a filter to filter elements according to a string attribute using
        wildcard expressions.

        :param lhs: the field to compare
        :type lhs: :class:`django.db.models.F`
        :param rhs: the wildcard pattern: a string containing any number of '%'
                    characters as wildcards.
        :type rhs: str
        :param case: whether the lookup shall be done case sensitively or not
        :type case: bool
        :param not_: whether the range shall be inclusive (the default) or
                     exclusive
        :type not_: bool
        :param mapping_choices: a dict to lookup potential choices for a
                                certain field.
        :type mapping_choices: dict[str, str]
        :return: a comparison expression object
        :rtype: :class:`django.db.models.Q`
    """
    parts = pattern.split("%")
    length = len(parts)

    if mapping_choices and lhs.name in mapping_choices:
        # special case when choices are given for the field:
        # compare statically and use 'in' operator to check if contained
        cmp_av = [
            (a, a.lower() if nocase else a)
            for a in mapping_choices[lhs.name].keys()
        ]

        for idx, part in enumerate(parts):
            if not part:
                continue

            cmp_p = part.lower() if nocase else part

            if idx == 0 and length > 1:  # startswith
                cmp_av = [a for a in cmp_av if a[1].startswith(cmp_p)]
            elif idx == 0:  # exact matching
                cmp_av = [a for a in cmp_av if a[1] == cmp_p]
            elif idx == length - 1:   # endswith
                cmp_av = [a for a in cmp_av if a[1].endswith(cmp_p)]
            else:  # middle
                cmp_av = [a for a in cmp_av if cmp_p in a[1]]

        q = Q(**{
            "%s__in" % lhs.name: [
                mapping_choices[lhs.name][a[0]]
                for a in cmp_av
            ]
        })

    else:
        i = "i" if nocase else ""
        q = None

        for idx, part in enumerate(parts):
            if not part:
                continue

            if idx == 0 and length > 1:  # startswith
                new_q = Q(**{
                    "%s__%s" % (lhs.name, "%sstartswith" % i): part
                })
            elif idx == 0:  # exact matching
                new_q = Q(**{
                    "%s__%s" % (lhs.name, "%sexact" % i): part
                })
            elif idx == length - 1:   # endswith
                new_q = Q(**{
                    "%s__%s" % (lhs.name, "%sendswith" % i): part
                })
            else:  # middle
                new_q = Q(**{
                    "%s__%s" % (lhs.name, "%scontains" % i): part
                })

            q = q & new_q if q else new_q

    return ~q if not_ else q


def contains(lhs: F, items: List[Union[F, Value]], not_: bool = False,
             mapping_choices: Optional[Dict[str, Dict[str, str]]] = None) -> Q:
    """ Create a filter to match elements attribute to be in a list of choices.

        :param lhs: the field to compare
        :type lhs: :class:`django.db.models.F`
        :param items: a list of choices
        :type items: list
        :param not_: whether the range shall be inclusive (the default) or
                     exclusive
        :type not_: bool
        :param mapping_choices: a dict to lookup potential choices for a
                                certain field.
        :type mapping_choices: dict[str, str]
        :return: a comparison expression object
        :rtype: :class:`django.db.models.Q`
    """

    if mapping_choices is not None and lhs.name in mapping_choices:
        def map_value(item: Union[str, Value],
                      choices: Dict[str, str]) -> Union[str, Value]:
            try:
                if isinstance(item, str):
                    item = choices[item]
                elif isinstance(item, Value):
                    item = Value(choices[item.value])

            except KeyError as e:
                raise AssertionError("Invalid field value %s" % e)
            return item

        items = [
            map_value(item, mapping_choices[lhs.name])
            for item in items
        ]

    q = Q(**{"%s__in" % lhs.name: items})
    return ~q if not_ else q


def null(lhs: F, not_: bool = False) -> Q:
    """ Create a filter to match elements whose attribute is (not) null

        :param lhs: the field to compare
        :type lhs: :class:`django.db.models.F`
        :param not_: whether the range shall be inclusive (the default) or
                     exclusive
        :type not_: bool
        :return: a comparison expression object
        :rtype: :class:`django.db.models.Q`
    """
    return Q(**{"%s__isnull" % lhs.name: not not_})


def temporal(lhs: F, time_or_period: Value, op: str) -> Q:
    """ Create a temporal filter for the given temporal attribute.

        :param lhs: the field to compare
        :type lhs: :class:`django.db.models.F`
        :param time_or_period: the time instant or time span to use as a filter
        :type time_or_period: :class:`datetime.datetime` or a tuple of two
                              datetimes or a tuple of one datetime and one
                              :class:`datetime.timedelta`
        :param op: the comparison operation. one of ``"BEFORE"``,
                   ``"BEFORE OR DURING"``, ``"DURING"``, ``"DURING OR AFTER"``,
                   ``"AFTER"``.
        :type op: str
        :return: a comparison expression object
        :rtype: :class:`django.db.models.Q`
    """
    assert op in (
        "BEFORE", "BEFORE OR DURING", "DURING", "DURING OR AFTER", "AFTER"
    )
    time_or_period = time_or_period.value
    low: Union[datetime, timedelta, None] = None
    high: Union[datetime, timedelta, None] = None
    if op in ("BEFORE", "AFTER"):
        assert isinstance(time_or_period, datetime)
        if op == "BEFORE":
            high = time_or_period
        else:
            low = time_or_period
    else:
        low, high = time_or_period
        low = low.value if isinstance(low, Value) else low
        high = high.value if isinstance(high, Value) else high
        assert isinstance(low, datetime) or isinstance(high, datetime)

        if isinstance(low, timedelta) and isinstance(high, datetime):
            low = high - low
        if isinstance(low, datetime) and isinstance(high, timedelta):
            high = low + high

    if low and high:
        return Q(**{"%s__range" % lhs.name: (low, high)})
    elif low:
        return Q(**{"%s__gte" % lhs.name: low})
    else:
        return Q(**{"%s__lte" % lhs.name: high})


def time_interval(time_or_period: Value, containment: str = 'overlaps',
                  begin_time_field: str = 'begin_time',
                  end_time_field: str = 'end_time') -> Q:
    """
    """

    gt_op = "__gte"
    lt_op = "__lte"

    is_slice = len(time_or_period) == 1
    if len(time_or_period) == 1:
        is_slice = True
        value = time_or_period[0]
    else:
        is_slice = False
        low, high = time_or_period

    if is_slice or (high == low and containment == "overlaps"):
        return Q(**{
            begin_time_field + "__lte": time_or_period[0],
            end_time_field + "__gte": time_or_period[0]
        })

    elif high == low:
        return Q(**{
            begin_time_field + "__gte": value,
            end_time_field + "__lte": value
        })

    else:
        q = Q()
        # check if the temporal bounds must be strictly contained
        if containment == "contains":
            if high is not None:
                q &= Q(**{
                    end_time_field + lt_op: high
                })
            if low is not None:
                q &= Q(**{
                    begin_time_field + gt_op: low
                })
        # or just overlapping
        else:
            if high is not None:
                q &= Q(**{
                    begin_time_field + lt_op: high
                })
            if low is not None:
                q &= Q(**{
                    end_time_field + gt_op: low
                })
        return q


UNITS_LOOKUP = {
    "kilometers": "km",
    "meters": "m"
}


INVERT_SPATIAL_OP = {
    "WITHIN": "CONTAINS",
    "CONTAINS": "WITHIN",
}


def spatial(lhs: Union[F, Value], rhs: Union[F, Value], op: str,
            pattern: Optional[str] = None, distance: Optional[float] = None,
            units: Optional[str] = None) -> Q:
    """ Create a spatial filter for the given spatial attribute.

        :param lhs: the field to compare
        :type lhs: :class:`django.db.models.F`
        :param rhs: the time instant or time span to use as a filter
        :type rhs:
        :param op: the comparison operation. one of ``"INTERSECTS"``,
                   ``"DISJOINT"``, `"CONTAINS"``, ``"WITHIN"``,
                   ``"TOUCHES"``, ``"CROSSES"``, ``"OVERLAPS"``,
                   ``"EQUALS"``, ``"RELATE"``, ``"DWITHIN"``, ``"BEYOND"``
        :type op: str
        :param pattern: the spatial relation pattern
        :type pattern: str
        :param distance: the distance value for distance based lookups:
                         ``"DWITHIN"`` and ``"BEYOND"``
        :type distance: float
        :param units: the units the distance is expressed in
        :type units: str
        :return: a comparison expression object
        :rtype: :class:`django.db.models.Q`
    """

    assert op in (
        "INTERSECTS", "DISJOINT", "CONTAINS", "WITHIN", "TOUCHES", "CROSSES",
        "OVERLAPS", "EQUALS", "RELATE", "DWITHIN", "BEYOND"
    )

    # if the left hand side is not a field reference, the comparison
    # can be be inverted to try if the right hand side is a field
    # reference.
    if not isinstance(lhs, F):
        lhs, rhs = rhs, lhs
        op = INVERT_SPATIAL_OP.get(op, op)

    # if neither lhs and rhs are fields, we have to fail here
    if not isinstance(lhs, F):
        raise ValueError(f'Unable to compare non-field {lhs}')

    return Q(**{"%s__%s" % (lhs.name, op.lower()): rhs})


def spatial_relate(lhs: Union[F, Value], rhs: Union[F, Value],
                   pattern: str) -> Q:

    if not isinstance(lhs, F):
        # TODO: cannot yet invert pattern -> raise
        raise ValueError(f'Unable to compare non-field {lhs}')

    return Q(**{"%s__relate" % lhs.name: (rhs, pattern)})


def spatial_distance(lhs: Union[F, Value], rhs: Union[F, Value], op: str,
                     distance: float, units: str) -> Q:
    if not isinstance(lhs, F):
        lhs, rhs = rhs, lhs

    # if neither lhs and rhs are fields, we have to fail here
    if not isinstance(lhs, F):
        raise ValueError(f'Unable to compare non-field {lhs}')

    # TODO: maybe use D.unit_attname(units)
    d = D(**{UNITS_LOOKUP[units]: distance})
    if op == "DWITHIN":
        return Q(**{"%s__distance_lte" % lhs.name: (rhs, d, 'spheroid')})
    return Q(**{"%s__distance_gte" % lhs.name: (rhs, d, 'spheroid')})


def bbox(lhs: F, minx: float, miny: float, maxx, maxy: float,
         crs: Optional[str] = None, bboverlaps: bool = True) -> Q:
    """ Create a bounding box filter for the given spatial attribute.

        :param lhs: the field to compare
        :param minx: the lower x part of the bbox
        :type minx: float
        :param miny: the lower y part of the bbox
        :type miny: float
        :param maxx: the upper x part of the bbox
        :type maxx: float
        :param maxy: the upper y part of the bbox
        :type maxy: float
        :param crs: the CRS the bbox is expressed in
        :type crs: str
        :type lhs: :class:`django.db.models.F`
        :return: a comparison expression object
        :rtype: :class:`django.db.models.Q`
    """
    box = Polygon.from_bbox((minx, miny, maxx, maxy))

    if crs:
        box.srid = SpatialReference(crs).srid
        box.transform(4326)

    if bboverlaps:
        return Q(**{"%s__bboverlaps" % lhs.name: box})
    return Q(**{"%s__intersects" % lhs.name: box})


def attribute(name: str, field_mapping: Optional[Dict[str, str]] = None) -> F:
    """ Create an attribute lookup expression using a field mapping dictionary.

        :param name: the field filter name
        :type name: str
        :param field_mapping: the dictionary to use as a lookup.
        :rtype: :class:`django.db.models.F`
    """
    if field_mapping:
        field = field_mapping.get(name, name)
    else:
        field = name
    return F(field)


def literal(value) -> Value:
    return Value(value)


OP_TO_FUNC = {
    "+": add,
    "-": sub,
    "*": mul,
    "/": truediv
}


def arithmetic(lhs: ArithmeticType, rhs: ArithmeticType,
               op: str) -> ArithmeticType:
    """ Create an arithmetic filter

        :param lhs: left hand side of the arithmetic expression. either a
                    scalar or a field lookup or another type of expression
        :param rhs: same as `lhs`
        :param op: the arithmetic operation. one of ``"+"``, ``"-"``, ``"*"``,
                   ``"/"``
        :rtype: :class:`django.db.models.F`
    """
    func = OP_TO_FUNC[op]
    return func(lhs, rhs)



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from datetime import date, datetime, time, timedelta
import operator
from typing import Callable, Dict

import shapely

from .evaluator import Evaluator, handle
from .. import ast
from .. import values
from ..util import like_pattern_to_re


COMPARISON_MAP = {
    '=': operator.eq,
    '<>': operator.ne,
    '<': operator.lt,
    '<=': operator.le,
    '>': operator.gt,
    '>=': operator.ge,
}

ARITHMETIC_MAP = {
    '+': operator.add,
    '-': operator.sub,
    '*': operator.mul,
    '/': operator.truediv,
}


def is_literal(value):
    return isinstance(value, values.LITERALS)


TEMPORAL_LITERALS = (date, datetime, time, timedelta, values.Interval)


def is_temporal_literal(value):
    return isinstance(value, TEMPORAL_LITERALS)


GEOMETRY_LITERALS = (values.Geometry, values.Envelope)


def is_geometry_literal(value):
    return isinstance(value, GEOMETRY_LITERALS)


def is_any_literal(value):
    return (
        is_literal(value)
        or is_temporal_literal(value)
        or is_geometry_literal(value)
    )


def to_geometry(value):
    if isinstance(value, values.Geometry):
        return shapely.geometry.shape(value)
    elif isinstance(value, values.Envelope):
        return shapely.geometry.Polygon.from_bounds(
            value.x1, value.y1, value.x2, value.y2
        )
    raise ValueError(str(type(value)))


class OptimizeEvaluator(Evaluator):
    def __init__(self, function_map: Dict[str, Callable]):
        self.function_map = function_map

    @handle(ast.Not)
    def not_(self, node, sub):
        if isinstance(sub, bool):
            return operator.not_(sub)
        else:
            return ast.Not(sub)

    @handle(ast.And, ast.Or)
    def combination(self, node, lhs, rhs):
        if isinstance(lhs, bool) and isinstance(rhs, bool):
            op = operator.and_ if node.op.value == 'AND' else operator.or_
            return op(lhs, rhs)

        elif isinstance(lhs, bool) or isinstance(rhs, bool):
            if isinstance(lhs, bool):
                certain, uncertain = lhs, rhs
            else:
                certain, uncertain = rhs, lhs

            # for OR nodes, when we have one true branch, the other
            # can be dropped. Otherwise we can shorthand to the
            # uncertain branch
            if node.op.value == 'OR':
                if certain:
                    return True
                else:
                    return uncertain
            # for AND nodes, we can drop the node if the certain one is
            # false. Otherwise we can shorthand to the other
            elif node.op.value == 'AND':
                if certain:
                    return uncertain
                else:
                    return False
            # we can eliminate the whole node and its sub-nodes, as it
            # will always evaluate to false
            return False

        else:
            return (ast.And if node.op.value == 'AND' else ast.Or)(lhs, rhs)

    @handle(ast.Comparison, subclasses=True)
    def comparison(self, node, lhs, rhs):
        if is_literal(lhs) and is_literal(rhs):
            op = COMPARISON_MAP[node.op.value]
            return op(lhs, rhs)
        else:
            return type(node)(lhs, rhs)

    @handle(ast.Between)
    def between(self, node, lhs, low, high):
        if is_literal(lhs) and is_literal(low) and is_literal(high):
            result = low <= lhs <= high
            if node.not_:
                result = not result
            return result
        else:
            return ast.Between(
                lhs,
                low,
                high,
                node.not_
            )

    @handle(ast.Like)
    def like(self, node, lhs):
        if is_literal(lhs):
            regex = like_pattern_to_re(
                node.pattern,
                node.nocase,
                node.wildcard,
                node.singlechar,
                node.escapechar
            )
            result = regex.match(lhs) is not None
            if node.not_:
                result = not result
            return result
        else:
            return ast.Like(
                lhs,
                node.pattern,
                node.nocase,
                node.wildcard,
                node.singlechar,
                node.escapechar,
                node.not_,
            )

    @handle(ast.In)
    def in_(self, node, lhs, *options):
        if is_literal(lhs) and all(is_literal(o) for o in options):
            result = lhs in options
            if node.not_:
                result = not result
            return result
        else:
            return ast.In(lhs, list(options), node.not_)

    @handle(ast.IsNull)
    def null(self, node, lhs):
        return ast.IsNull(lhs, node.not_)

    @handle(ast.Exists)
    def exists(self, node, lhs):
        return ast.Exists(lhs, node.not_)

    @handle(ast.TemporalPredicate, subclasses=True)
    def temporal(self, node, lhs, rhs):
        if is_temporal_literal(lhs) and is_temporal_literal(rhs):
            lhs = to_interval(lhs)
            rhs = to_interval(rhs)

            return node.op.value == relate_intervals(lhs, rhs)
        else:
            return type(node)(lhs, rhs)

    @handle(ast.ArrayPredicate, subclasses=True)
    def array(self, node, lhs, rhs):
        if isinstance(lhs, list) and isinstance(rhs, list):
            left = set(lhs)
            right = set(rhs)

            if node.op == ast.ArrayComparisonOp.AEQUALS:
                return left == right
            elif node.op == ast.ArrayComparisonOp.ACONTAINS:
                return left >= right
            elif node.op == ast.ArrayComparisonOp.ACONTAINEDBY:
                return left <= right
            elif node.op == ast.ArrayComparisonOp.AOVERLAPS:
                return bool(left & right)
        else:
            return type(node)(lhs, rhs)

    @handle(ast.SpatialComparisonPredicate, subclasses=True)
    def spatial_operation(self, node, lhs, rhs):
        if is_geometry_literal(lhs) and is_geometry_literal(rhs):
            lhs = to_geometry(lhs)
            rhs = to_geometry(rhs)
            op = getattr(lhs, node.op.value.lower())
            return op(rhs)
        else:
            return type(node)(lhs, rhs)

    @handle(ast.Relate)
    def spatial_pattern(self, node, lhs, rhs):
        if is_geometry_literal(lhs) and isinstance(rhs, str):
            lhs = to_geometry(lhs)
            return lhs.relate_pattern(rhs, node.pattern)
        else:
            return ast.Relate(lhs, rhs, node.pattern)

    @handle(ast.SpatialDistancePredicate)
    def distance(self, node, lhs, rhs):
        # TODO: can this be reduced?
        return type(node)(lhs, rhs, node.distance, node.units)

    @handle(ast.BBox)
    def bbox(self, node, lhs):
        if is_geometry_literal(lhs):
            lhs = to_geometry(lhs)
            return lhs.intersects(
                shapely.geometry.Polygon.from_bounds(
                    node.minx, node.miny, node.maxx, node.maxy
                )
            )
        else:
            return ast.BBox(lhs, node.minx, node.miny, node.maxx, node.maxy)

    @handle(ast.Attribute)
    def attribute(self, node):
        return node

    @handle(ast.Arithmetic, subclasses=True)
    def arithmetic(self, node, lhs, rhs):
        if is_literal(lhs) and is_literal(rhs):
            op = ARITHMETIC_MAP[node.op.value]
            return op(lhs, rhs)
        else:
            return type(node)(lhs, rhs)

    @handle(ast.Function)
    def function(self, node, *arguments):
        func = self.function_map.get(node.name)
        if func and all(is_any_literal(a) for a in arguments):
            return func(*arguments)
        else:
            return ast.Function(node.name, list(arguments))

    # just pass through these nodes
    @handle(ast.Attribute, values.Geometry, values.Envelope, *values.LITERALS)
    def literal(self, node):
        return node


def to_interval(value):
    # TODO:
    zulu = None
    if isinstance(value, values.Interval):
        low = value.start
        high = value.end
        if isinstance(low, date):
            low = datetime.combine(low, time.min, zulu)
        if isinstance(high, date):
            high = datetime.combine(high, time.max, zulu)

        if isinstance(low, timedelta):
            low = high - timedelta
        elif isinstance(high, timedelta):
            high = low + timedelta

        return (low, high)

    elif isinstance(value, date):
        return (
            datetime.combine(value, time.min, zulu),
            datetime.combine(value, time.max, zulu),
        )

    elif isinstance(value, datetime):
        return (value, value)

    raise ValueError(f'Invalid type {type(value)}')


def relate_intervals(lhs, rhs):
    ll, lh = lhs
    rl, rh = rhs
    if lh < rl:
        return 'BEFORE'
    elif ll > rh:
        return 'AFTER'
    elif lh == rl:
        return 'MEETS'
    elif ll == rh:
        return 'METBY'
    elif ll < rl and rl < lh < rh:
        return 'TOVERLAPS'
    elif rl < ll < rh and lh > rh:
        return 'OVERLAPPEDBY'
    elif ll == rl and lh < rh:
        return 'BEGINS'
    elif ll == rl and lh > rh:
        return 'BEGUNBY'
    elif ll > rl and lh < rh:
        return 'DURING'
    elif ll < rl and lh > rh:
        return 'TCONTAINS'
    elif ll > rl and lh == rh:
        return 'TENDS'
    elif ll < rl and lh == rh:
        return 'ENDEDBY'
    elif ll == rl and lh == rh:
        return 'TEQUALS'

    raise ValueError(
        f'Error relating intervals [{ll}, {lh}] and ({rl}, {rh})'
    )


def optimize(
        root: ast.Node,
        function_map: Dict[str, Callable] = None) -> ast.Node:
    result = OptimizeEvaluator(function_map or {}).evaluate(root)
    if isinstance(result, bool):
        result = ast.Include(not result)

    return result




from datetime import datetime, date, time, timedelta

from . import filters
from ... import ast
from ... import values
from ..evaluator import Evaluator, handle


LITERALS = (str, float, int, bool, datetime, date, time, timedelta)


class SQLAlchemyFilterEvaluator(Evaluator):
    def __init__(self, field_mapping):
        self.field_mapping = field_mapping

    @handle(ast.Not)
    def not_(self, node, sub):
        return filters.negate(sub)

    @handle(ast.And, ast.Or)
    def combination(self, node, lhs, rhs):
        return filters.combine((lhs, rhs), node.op.value)

    @handle(ast.Comparison, subclasses=True)
    def comparison(self, node, lhs, rhs):
        return filters.runop(
            lhs,
            rhs,
            node.op.value,
        )

    @handle(ast.Between)
    def between(self, node, lhs, low, high):
        return filters.between(
            lhs,
            low,
            high,
            node.not_
        )

    @handle(ast.Like)
    def like(self, node, lhs):
        return filters.like(
            lhs,
            node.pattern,
            not node.nocase,
            node.not_,
        )

    @handle(ast.In)
    def in_(self, node, lhs, *options):
        return filters.runop(
            lhs,
            options,
            "in",
            node.not_,
        )

    @handle(ast.IsNull)
    def null(self, node, lhs):
        return filters.runop(
            lhs,
            None,
            "is_null",
            node.not_
        )

    # @handle(ast.ExistsPredicateNode)
    # def exists(self, node, lhs):
    #     if self.use_getattr:
    #         result = hasattr(self.obj, node.lhs.name)
    #     else:
    #         result = lhs in self.obj

    #     if node.not_:
    #         result = not result
    #     return result

    @handle(ast.TemporalPredicate, subclasses=True)
    def temporal(self, node, lhs, rhs):
        return filters.temporal(
            lhs,
            rhs,
            node.op.value,
        )

    @handle(ast.SpatialComparisonPredicate, subclasses=True)
    def spatial_operation(self, node, lhs, rhs):
        return filters.spatial(
            lhs,
            rhs,
            node.op.name,
        )

    @handle(ast.Relate)
    def spatial_pattern(self, node, lhs, rhs):
        return filters.spatial(
            lhs,
            rhs,
            'RELATE',
            pattern=node.pattern,
        )

    @handle(ast.SpatialDistancePredicate, subclasses=True)
    def spatial_distance(self, node, lhs, rhs):
        return filters.spatial(
            lhs,
            rhs,
            node.op.value,
            distance=node.distance,
            units=node.units,
        )

    @handle(ast.BBox)
    def bbox(self, node, lhs):
        return filters.bbox(
            lhs,
            node.minx,
            node.miny,
            node.maxx,
            node.maxy,
            node.crs
        )

    @handle(ast.Attribute)
    def attribute(self, node):
        return filters.attribute(
            node.name,
            self.field_mapping
        )

    @handle(ast.Arithmetic, subclasses=True)
    def arithmetic(self, node, lhs, rhs):
        return filters.runop(
            lhs,
            rhs,
            node.op.value
        )

    # TODO: map functions
    # @handle(ast.FunctionExpressionNode)
    # def function(self, node, *arguments):
    #     return self.function_map[node.name](*arguments)

    @handle(*values.LITERALS)
    def literal(self, node):
        return filters.literal(node)

    @handle(values.Interval)
    def interval(self, node, start, end):
        return filters.literal((start, end))

    @handle(values.Geometry)
    def geometry(self, node):
        return filters.parse_geometry(node.__geo_interface__)

    @handle(values.Envelope)
    def envelope(self, node):
        return filters.parse_bbox([node.x1, node.y1, node.x2, node.y2])


def to_filter(ast, field_mapping=None):
    """ Helper function to translate ECQL AST to Django Query expressions.

        :param ast: the abstract syntax tree
        :param field_mapping: a dict mapping from the filter name to the Django
                              field lookup.
        :param mapping_choices: a dict mapping field lookups to choices.
        :type ast: :class:`Node`
        :returns: a Django query object
        :rtype: :class:`django.db.models.Q`
    """
    return SQLAlchemyFilterEvaluator(field_mapping).evaluate(ast)



import base64
import datetime

from ... import ast
from ...util import parse_datetime, parse_duration
from .util import handle, ParseInput, Element
from .base import FESBaseParser


class FES20Parser(FESBaseParser):
    namespace = 'http://www.opengis.net/fes/2.0'

    # @handle('PropertyIsNil')
    # def property_is_nil(self, node: Element, lhs, rhs):
    #     return ast...

    @handle('After')
    def time_after(self, node: Element, lhs, rhs):
        return ast.TimeAfter(lhs, rhs)

    @handle('Before')
    def time_before(self, node: Element, lhs, rhs):
        return ast.TimeBefore(lhs, rhs)

    @handle('Begins')
    def time_begins(self, node: Element, lhs, rhs):
        return ast.TimeBegins(lhs, rhs)

    @handle('BegunBy')
    def time_begun_by(self, node: Element, lhs, rhs):
        return ast.TimeBegunBy(lhs, rhs)

    @handle('TContains')
    def time_contains(self, node: Element, lhs, rhs):
        return ast.TimeContains(lhs, rhs)

    @handle('During')
    def time_during(self, node: Element, lhs, rhs):
        return ast.TimeDuring(lhs, rhs)

    @handle('TEquals')
    def time_equals(self, node: Element, lhs, rhs):
        return ast.TimeEquals(lhs, rhs)

    @handle('TOverlaps')
    def time_overlaps(self, node: Element, lhs, rhs):
        return ast.TimeOverlaps(lhs, rhs)

    @handle('Meets')
    def time_meets(self, node: Element, lhs, rhs):
        return ast.TimeMeets(lhs, rhs)

    @handle('OverlappedBy')
    def time_overlapped_by(self, node: Element, lhs, rhs):
        return ast.TimeOverlappedBy(lhs, rhs)

    @handle('MetBy')
    def time_met_by(self, node: Element, lhs, rhs):
        return ast.TimeMetBy(lhs, rhs)

    @handle('Ends')
    def time_ends(self, node: Element, lhs, rhs):
        return ast.TimeEnds(lhs, rhs)

    @handle('EndedBy')
    def time_ended_by(self, node: Element, lhs, rhs):
        return ast.TimeEndedBy(lhs, rhs)

    @handle('ValueReference')
    def value_reference(self, node: Element):
        return ast.Attribute(node.text)

    @handle('Literal')
    def literal(self, node: Element):
        type_ = node.get('type').rpartition(':')[2]
        value = node.text
        if type_ == 'boolean':
            return value.lower() == 'true'
        elif type_ in ('byte', 'int', 'integer', 'long', 'negativeInteger',
                       'nonNegativeInteger', 'nonPositiveInteger',
                       'positiveInteger', 'short', 'unsignedByte',
                       'unsignedInt', 'unsignedLong', 'unsignedShort'):
            return int(value)
        elif type_ in ('decimal', 'double', 'float'):
            return float(value)
        elif type_ == 'base64Binary':
            return base64.b64decode(value)
        elif type_ == 'hexBinary':
            return bytes.fromhex(value)
        elif type_ == 'date':
            return datetime.date.fromisoformat(value)
        elif type_ == 'dateTime':
            return parse_datetime(value)
        elif type_ == 'duration':
            return parse_duration(value)

        # return to string
        return value


def parse(input_: ParseInput) -> ast.Node:
    return FES20Parser().parse(input_)



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2019 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import json

from django.contrib.gis.geos import GEOSGeometry, Polygon

from . import filters
from ... import ast
from ... import values
from ..evaluator import Evaluator, handle


class DjangoFilterEvaluator(Evaluator):
    def __init__(self, field_mapping, mapping_choices):
        self.field_mapping = field_mapping
        self.mapping_choices = mapping_choices

    @handle(ast.Not)
    def not_(self, node, sub):
        return filters.negate(sub)

    @handle(ast.And, ast.Or)
    def combination(self, node, lhs, rhs):
        return filters.combine((lhs, rhs), node.op.value)

    @handle(ast.Comparison, subclasses=True)
    def comparison(self, node, lhs, rhs):
        return filters.compare(
            lhs,
            rhs,
            node.op.value,
            self.mapping_choices
        )

    @handle(ast.Between)
    def between(self, node, lhs, low, high):
        return filters.between(
            lhs,
            low,
            high,
            node.not_
        )

    @handle(ast.Like)
    def like(self, node, lhs):
        return filters.like(
            lhs,
            node.pattern,
            node.nocase,
            node.not_,
            self.mapping_choices
        )

    @handle(ast.In)
    def in_(self, node, lhs, *options):
        return filters.contains(
            lhs,
            options,
            node.not_,
            self.mapping_choices
        )

    @handle(ast.IsNull)
    def null(self, node, lhs):
        return filters.null(
            lhs,
            node.not_
        )

    # @handle(ast.ExistsPredicateNode)
    # def exists(self, node, lhs):
    #     if self.use_getattr:
    #         result = hasattr(self.obj, node.lhs.name)
    #     else:
    #         result = lhs in self.obj

    #     if node.not_:
    #         result = not result
    #     return result

    @handle(ast.TemporalPredicate, subclasses=True)
    def temporal(self, node, lhs, rhs):
        return filters.temporal(
            lhs,
            rhs,
            node.op.value,
        )

    @handle(ast.SpatialComparisonPredicate, subclasses=True)
    def spatial_operation(self, node, lhs, rhs):
        return filters.spatial(
            lhs,
            rhs,
            node.op.name,
        )

    @handle(ast.Relate)
    def spatial_pattern(self, node, lhs, rhs):
        return filters.spatial_relate(
            lhs,
            rhs,
            pattern=node.pattern,
        )

    @handle(ast.SpatialDistancePredicate, subclasses=True)
    def spatial_distance(self, node, lhs, rhs):
        return filters.spatial_distance(
            lhs,
            rhs,
            node.op.value,
            distance=node.distance,
            units=node.units,
        )

    @handle(ast.BBox)
    def bbox(self, node, lhs):
        return filters.bbox(
            lhs,
            node.minx,
            node.miny,
            node.maxx,
            node.maxy,
            node.crs
        )

    @handle(ast.Attribute)
    def attribute(self, node):
        return filters.attribute(node.name, self.field_mapping)

    @handle(ast.Arithmetic, subclasses=True)
    def arithmetic(self, node, lhs, rhs):
        return filters.arithmetic(
            lhs,
            rhs,
            node.op.value
        )

    # TODO: map functions
    # @handle(ast.FunctionExpressionNode)
    # def function(self, node, *arguments):
    #     return self.function_map[node.name](*arguments)

    @handle(*values.LITERALS)
    def literal(self, node):
        return filters.literal(node)

    @handle(values.Interval)
    def interval(self, node, start, end):
        return filters.literal((start, end))

    @handle(values.Geometry)
    def geometry(self, node):
        return GEOSGeometry(json.dumps(node.__geo_interface__))

    @handle(values.Envelope)
    def envelope(self, node):
        return Polygon.from_bbox(
            (node.x1, node.y1, node.x2, node.y2)
        )


def to_filter(root, field_mapping=None, mapping_choices=None):
    """ Helper function to translate ECQL AST to Django Query expressions.

        :param ast: the abstract syntax tree
        :param field_mapping: a dict mapping from the filter name to the Django
                              field lookup.
        :param mapping_choices: a dict mapping field lookups to choices.
        :type ast: :class:`Node`
        :returns: a Django query object
        :rtype: :class:`django.db.models.Q`
    """
    return DjangoFilterEvaluator(
        field_mapping, mapping_choices
    ).evaluate(root)



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from lark import Transformer, v_args

from ..util import parse_duration, parse_datetime


@v_args(inline=True)
class ISO8601Transformer(Transformer):
    def DATETIME(self, dt):
        return parse_datetime(dt)

    def DURATION(self, duration):
        return parse_duration(duration)



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2019 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import re
from datetime import timedelta, date, datetime
from dateparser import parse as _parse_datetime

__all__ = [
    'parse_datetime', 'RE_ISO_8601', 'parse_duration',
    'like_pattern_to_re_pattern', 'like_pattern_to_re'
]

RE_ISO_8601 = re.compile(
    r"^(?P<sign>[+-])?P"
    r"(?:(?P<years>\d+(\.\d+)?)Y)?"
    r"(?:(?P<months>\d+(\.\d+)?)M)?"
    r"(?:(?P<days>\d+(\.\d+)?)D)?"
    r"T?(?:(?P<hours>\d+(\.\d+)?)H)?"
    r"(?:(?P<minutes>\d+(\.\d+)?)M)?"
    r"(?:(?P<seconds>\d+(\.\d+)?)S)?$"
)


def parse_duration(value: str) -> timedelta:
    """ Parses an ISO 8601 duration string into a python timedelta object.
        Raises a ``ValueError`` if a conversion was not possible.

        :param value: the ISO8601 duration string to parse
        :type value: str
        :return: the parsed duration
        :rtype: datetime.timedelta
    """

    match = RE_ISO_8601.match(value)
    if not match:
        raise ValueError(
            "Could not parse ISO 8601 duration from '%s'." % value
        )
    parts = match.groupdict()

    sign = -1 if "-" == parts['sign'] else 1
    days = float(parts['days'] or 0)
    days += float(parts['months'] or 0) * 30  # ?!
    days += float(parts['years'] or 0) * 365  # ?!
    fsec = float(parts['seconds'] or 0)
    fsec += float(parts['minutes'] or 0) * 60
    fsec += float(parts['hours'] or 0) * 3600

    return sign * timedelta(days, fsec)


def parse_date(value: str) -> date:
    """ Backport for `fromisoformat` for dates in Python 3.6
    """
    return date(*(int(part) for part in value.split('-')))


def parse_datetime(value: str) -> datetime:
    parsed = _parse_datetime(value)
    if parsed is None:
        raise ValueError(value)
    return parsed


def like_pattern_to_re_pattern(like, wildcard, single_char, escape_char):
    x_wildcard = re.escape(wildcard)
    x_single_char = re.escape(single_char)

    dx_wildcard = re.escape(x_wildcard)
    dx_single_char = re.escape(x_single_char)

    # special handling if escape char clashes with re escape char
    if escape_char == '\\':
        x_escape_char = '\\\\\\\\'
    else:
        x_escape_char = re.escape(escape_char)
    dx_escape_char = re.escape(x_escape_char)

    pattern = re.escape(like)

    # handle not escaped wildcards/single chars
    pattern = re.sub(
        f'(?<!{x_escape_char}){dx_wildcard}',
        '.*',
        pattern,
    )
    pattern = re.sub(
        f'(?<!{x_escape_char}){dx_single_char}',
        '.',
        pattern,
    )

    # handle escaped wildcard, single chars and escape chars
    pattern = re.sub(
        f'{dx_escape_char}{dx_wildcard}',
        x_wildcard,
        pattern,
    )
    pattern = re.sub(
        f'{dx_escape_char}{dx_single_char}',
        x_single_char,
        pattern,
    )
    pattern = re.sub(
        f'{x_escape_char}{x_escape_char}',
        x_escape_char,
        pattern,
    )

    return f'^{pattern}$'


def like_pattern_to_re(like, nocase, wildcard, single_char, escape_char):
    flags = re.I if nocase else 0
    return re.compile(
        like_pattern_to_re_pattern(like, wildcard, single_char, escape_char),
        flags=flags
    )



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from lark import Transformer, v_args


@v_args(inline=True)
class WKTTransformer(Transformer):
    def wkt__geometry_with_srid(self, srid, geometry):
        print(srid, geometry)
        geometry["crs"] = {
            "type": "name",
            "properties": {
                "name": f"urn:ogc:def:crs:EPSG::{srid}"
            }
        }
        return geometry

    def wkt__geometrycollection(self, *geometries):
        return {
            "type": "GeometryCollection",
            "geometries": geometries
        }

    def wkt__point(self, coordinates):
        return {
            "type": "Point",
            "coordinates": coordinates,
        }

    def wkt__linestring(self, coordinate_list):
        return {
            "type": "LineString",
            "coordinates": coordinate_list,
        }

    def wkt__polygon(self, coordinate_lists):
        return {
            "type": "Polygon",
            "coordinates": coordinate_lists,
        }

    def wkt__multipoint(self, coordinates):
        return {
            "type": "MultiPoint",
            "coordinates": coordinates,
        }

    def wkt__multipoint_2(self, *coordinates):
        print(coordinates)
        return {
            "type": "MultiPoint",
            "coordinates": coordinates,
        }

    def wkt__multilinestring(self, coordinate_lists):
        return {
            "type": "MultiLineString",
            "coordinates": coordinate_lists,
        }

    def wkt__multipolygon(self, *coordinate_lists):
        return {
            "type": "MultiPolygon",
            "coordinates": coordinate_lists,
        }

    def wkt__coordinate_lists(self, *coordinate_lists):
        return coordinate_lists

    def wkt__coordinate_list(self, coordinate_list, coordinate):
        return coordinate_list + (coordinate,)

    def wkt__coordinate_list_start(self, coordinate_list):
        return (coordinate_list,)

    def wkt__coordinate(self, *components):
        return components

    def wkt__SIGNED_NUMBER(self, value):
        return float(value)

    def wkt__NUMBER(self, value):
        return float(value)



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from functools import wraps
from typing import Any, Callable, Dict, List, Type, cast

from .. import ast


def get_all_subclasses(*classes: Type) -> List[Type]:
    """Utility function to get all the leaf-classes (classes that don't
    have any further sub-classes) from a given list of classes.
    """
    all_subclasses = []

    for cls in classes:
        subclasses = cls.__subclasses__()
        if subclasses:
            all_subclasses.extend(get_all_subclasses(*subclasses))
        else:
            # directly insert classes that do not have any sub-classes
            all_subclasses.append(cls)

    return all_subclasses


def handle(*node_classes: Type, subclasses: bool = False) -> Callable:
    """Function-decorator to mark a class function as a handler for a
    given node type.
    """
    assert node_classes

    @wraps(handle)
    def inner(func):
        if subclasses:
            func.handles_classes = get_all_subclasses(*node_classes)
        else:
            func.handles_classes = node_classes
        return func

    return inner


class EvaluatorMeta(type):
    """Metaclass for the ``Evaluator`` class to create a static map for
    all handler methods by their respective handled types.
    """

    def __init__(cls, name, bases, dct):
        cls.handler_map = {}
        for base in bases:
            cls.handler_map.update(getattr(base, "handler_map"))

        for value in dct.values():
            if hasattr(value, "handles_classes"):
                for handled_class in value.handles_classes:
                    cls.handler_map[handled_class] = value


class Evaluator(metaclass=EvaluatorMeta):
    """Base class for AST evaluators."""

    handler_map: Dict[Type, Callable]

    def evaluate(self, node: ast.AstType, adopt_result: bool = True) -> Any:
        """Recursive function to evaluate an abstract syntax tree.
        For every node in the walked syntax tree, its registered handler
        is called with the node as first parameter and all pre-evaluated
        child nodes as star-arguments.
        When no handler was found for a given node, the ``adopt`` function
        is called with the node and its arguments, which by default raises
        an ``NotImplementedError``.
        """
        sub_args = []
        if hasattr(node, "get_sub_nodes"):
            subnodes = cast(ast.Node, node).get_sub_nodes()
            if subnodes:
                if isinstance(subnodes, list):
                    sub_args = [
                        self.evaluate(sub_node, False) for sub_node in subnodes
                    ]
                else:
                    sub_args = [self.evaluate(subnodes, False)]

        handler = self.handler_map.get(type(node))
        if handler is not None:
            result = handler(self, node, *sub_args)
        else:
            result = self.adopt(node, *sub_args)

        if adopt_result:
            return self.adopt_result(result)
        else:
            return result

    def adopt(self, node, *sub_args):
        """Interface function for a last resort when trying to evaluate a node
        and no handler was found.
        """
        raise NotImplementedError(
            f"Failed to evaluate node of type {type(node)}"
        )

    def adopt_result(self, result: Any) -> Any:
        """Interface function for adopting the final evaluation result if necessary.
        Default is no-op.
        """
        return result



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from typing import Dict, Optional

import shapely.geometry

from ..evaluator import Evaluator, handle
from ... import ast
from ... import values


COMPARISON_OP_MAP = {
    ast.ComparisonOp.EQ: '=',
    ast.ComparisonOp.NE: '<>',
    ast.ComparisonOp.LT: '<',
    ast.ComparisonOp.LE: '<=',
    ast.ComparisonOp.GT: '>',
    ast.ComparisonOp.GE: '>=',
}


ARITHMETIC_OP_MAP = {
    ast.ArithmeticOp.ADD: '+',
    ast.ArithmeticOp.SUB: '-',
    ast.ArithmeticOp.MUL: '*',
    ast.ArithmeticOp.DIV: '/',
}

SPATIAL_COMPARISON_OP_MAP = {
    ast.SpatialComparisonOp.INTERSECTS: 'ST_Intersects',
    ast.SpatialComparisonOp.DISJOINT: 'ST_Disjoint',
    ast.SpatialComparisonOp.CONTAINS: 'ST_Contains',
    ast.SpatialComparisonOp.WITHIN: 'ST_Within',
    ast.SpatialComparisonOp.TOUCHES: 'ST_Touches',
    ast.SpatialComparisonOp.CROSSES: 'ST_Crosses',
    ast.SpatialComparisonOp.OVERLAPS: 'ST_Overlaps',
    ast.SpatialComparisonOp.EQUALS: 'ST_Equals',
}


class SQLEvaluator(Evaluator):
    def __init__(self, attribute_map: Dict[str, str],
                 function_map: Dict[str, str]):
        self.attribute_map = attribute_map
        self.function_map = function_map

    @handle(ast.Not)
    def not_(self, node, sub):
        return f"NOT {sub}"

    @handle(ast.And, ast.Or)
    def combination(self, node, lhs, rhs):
        return f'({lhs} {node.op.value} {rhs})'

    @handle(ast.Comparison, subclasses=True)
    def comparison(self, node, lhs, rhs):
        return f"({lhs} {COMPARISON_OP_MAP[node.op]} {rhs})"

    @handle(ast.Between)
    def between(self, node, lhs, low, high):
        return f"({lhs} {'NOT ' if node.not_ else ''}BETWEEN {low} AND {high})"

    @handle(ast.Like)
    def like(self, node, lhs):
        pattern = node.pattern
        if node.wildcard != '%':
            # TODO: not preceded by escapechar
            pattern = pattern.replace(node.wildcard, '%')
        if node.singlechar != '_':
            # TODO: not preceded by escapechar
            pattern = pattern.replace(node.singlechar, '_')

        # TODO: handle node.nocase
        return (
            f"{lhs} {'NOT ' if node.not_ else ''}LIKE "
            f"'{pattern}' ESCAPE '{node.escapechar}'"
        )

    @handle(ast.In)
    def in_(self, node, lhs, *options):
        return f"{lhs} {'NOT ' if node.not_ else ''}IN ({', '.join(options)})"

    @handle(ast.IsNull)
    def null(self, node, lhs):
        return f"{lhs} IS {'NOT ' if node.not_ else ''}NULL"

    # @handle(ast.TemporalPredicate, subclasses=True)
    # def temporal(self, node, lhs, rhs):
    #     pass

    @handle(ast.SpatialComparisonPredicate, subclasses=True)
    def spatial_operation(self, node, lhs, rhs):
        func = SPATIAL_COMPARISON_OP_MAP[node.op]
        return f"{func}({lhs},{rhs})"

    @handle(ast.BBox)
    def bbox(self, node, lhs):
        func = SPATIAL_COMPARISON_OP_MAP[ast.SpatialComparisonOp.INTERSECTS]
        # TODO: create BBox geometry
        rhs = ""
        return f"{func}({lhs},{rhs})"

    @handle(ast.Attribute)
    def attribute(self, node: ast.Attribute):
        return f"\"{self.attribute_map[node.name]}\""

    @handle(ast.Arithmetic, subclasses=True)
    def arithmetic(self, node: ast.Arithmetic, lhs, rhs):
        op = ARITHMETIC_OP_MAP[node.op]
        return f"({lhs} {op} {rhs})"

    @handle(ast.Function)
    def function(self, node, *arguments):
        func = self.function_map[node.name]
        return f"{func}({','.join(arguments)})"

    @handle(*values.LITERALS)
    def literal(self, node):
        if isinstance(node, str):
            return f"'{node}'"
        else:
            # TODO:
            return str(node)

    @handle(values.Geometry)
    def geometry(self, node: values.Geometry):
        wkb_hex = shapely.geometry.shape(node).wkb_hex
        return f"ST_GeomFromWKB(x'{wkb_hex}')"

    @handle(values.Envelope)
    def envelope(self, node: values.Envelope):
        wkb_hex = shapely.geometry.box(
            node.x1, node.y1, node.x2, node.y2
        ).wkb_hex
        return f"ST_GeomFromWKB(x'{wkb_hex}')"


def to_sql_where(root: ast.Node, field_mapping: Dict[str, str],
                 function_map: Optional[Dict[str, str]] = None) -> str:
    return SQLEvaluator(field_mapping, function_map or {}).evaluate(root)



import base64
import datetime

from pygml.v32 import (
    parse_v32, NAMESPACE as NAMESPACE_32, NSMAP as NSMAP_32
)
from pygml.pre_v32 import (
    parse_pre_v32, NAMESPACE as NAMESPACE_PRE_32, NSMAP as NSMAP_PRE_32
)
from pygml.v33 import parse_v33_ce, NAMESPACE as NAMESPACE_33_CE
from pygml.georss import NAMESPACE as NAMESPACE_GEORSS, parse_georss

from ... import ast
from ... import values
from ...util import parse_datetime, parse_duration
from .util import XMLParser, handle, handle_namespace, Element
from .gml import is_temporal, parse_temporal


class FESBaseParser(XMLParser):
    @handle('Filter')
    def filter_(self, node: Element, predicate):
        return predicate

    @handle('And')
    def and_(self, node: Element, lhs, rhs):
        return ast.And(lhs, rhs)

    @handle('Or')
    def or_(self, node: Element, lhs, rhs):
        return ast.Or(lhs, rhs)

    @handle('Not')
    def not_(self, node: Element, lhs):
        return ast.Not(lhs)

    @handle('PropertyIsEqualTo')
    def property_is_equal_to(self, node: Element, lhs, rhs):
        return ast.Equal(lhs, rhs)

    @handle('PropertyIsNotEqualTo')
    def property_is_not_equal_to(self, node: Element, lhs, rhs):
        return ast.NotEqual(lhs, rhs)

    @handle('PropertyIsLessThan')
    def property_is_less_than(self, node: Element, lhs, rhs):
        return ast.LessThan(lhs, rhs)

    @handle('PropertyIsGreaterThan')
    def property_is_greater_than(self, node: Element, lhs, rhs):
        return ast.GreaterThan(lhs, rhs)

    @handle('PropertyIsLessThanOrEqualTo')
    def property_is_less_than_or_equal_to(self, node: Element, lhs, rhs):
        return ast.LessEqual(lhs, rhs)

    @handle('PropertyIsGreaterThanOrEqualTo')
    def property_is_greater_than_or_equal_to(self, node: Element, lhs, rhs):
        return ast.GreaterEqual(lhs, rhs)

    @handle('PropertyIsLike')
    def property_is_like(self, node: Element, lhs, rhs):
        return ast.Like(
            lhs,
            rhs,
            wildcard=node.attrib['wildCard'],
            singlechar=node.attrib['singleChar'],
            escapechar=node.attrib.get('escape', node.attrib['escapeChar']),
            nocase=node.attrib.get('matchCase', 'true') == 'false',
            not_=False,
        )

    @handle('PropertyIsNull')
    def property_is_null(self, node: Element, lhs):
        return ast.IsNull(lhs, not_=False)

    @handle('PropertyIsBetween')
    def property_is_between(self, node: Element, lhs, low, high):
        return ast.Between(lhs, low, high, False)

    @handle('LowerBoundary', 'UpperBoundary')
    def boundary(self, node: Element, expression):
        return expression

    @handle('Equals')
    def geometry_equals(self, node: Element, lhs, rhs):
        return ast.GeometryEquals(lhs, rhs)

    @handle('Disjoint')
    def geometry_disjoint(self, node: Element, lhs, rhs):
        return ast.GeometryDisjoint(lhs, rhs)

    @handle('Touches')
    def geometry_touches(self, node: Element, lhs, rhs):
        return ast.GeometryTouches(lhs, rhs)

    @handle('Within')
    def geometry_within(self, node: Element, lhs, rhs):
        return ast.GeometryWithin(lhs, rhs)

    @handle('Overlaps')
    def geometry_overlaps(self, node: Element, lhs, rhs):
        return ast.GeometryOverlaps(lhs, rhs)

    @handle('Crosses')
    def geometry_crosses(self, node: Element, lhs, rhs):
        return ast.GeometryCrosses(lhs, rhs)

    @handle('Intersects')
    def geometry_intersects(self, node: Element, lhs, rhs):
        return ast.GeometryIntersects(lhs, rhs)

    @handle('Contains')
    def geometry_contains(self, node: Element, lhs, rhs):
        return ast.GeometryContains(lhs, rhs)

    @handle('DWithin')
    def distance_within(self, node: Element, lhs, rhs, distance_and_units):
        distance, units = distance_and_units
        return ast.DistanceWithin(lhs, rhs, distance, units)

    @handle('Beyond')
    def distance_beyond(self, node: Element, lhs, rhs, distance_and_units):
        distance, units = distance_and_units
        return ast.DistanceBeyond(lhs, rhs, distance, units)

    @handle('Distance')
    def distance(self, node: Element):
        return (float(node.text), node.attrib['uom'])

    # @handle('BBOX')
    # def geometry_bbox(self, node: Element, lhs, rhs):
    #     # TODO: ast.BBox() seems incompatible
    #     pass

    @handle('ValueReference')
    def value_reference(self, node):
        return ast.Attribute(node.text)

    @handle('Literal')
    def literal(self, node):
        type_ = node.get('type', '').rpartition(':')[2]
        value = node.text
        if type_ == 'boolean':
            return value.lower() == 'true'
        elif type_ in ('byte', 'int', 'integer', 'long', 'negativeInteger',
                       'nonNegativeInteger', 'nonPositiveInteger',
                       'positiveInteger', 'short', 'unsignedByte',
                       'unsignedInt', 'unsignedLong', 'unsignedShort'):
            return int(value)
        elif type_ in ('decimal', 'double', 'float'):
            return float(value)
        elif type_ == 'base64Binary':
            return base64.b64decode(value)
        elif type_ == 'hexBinary':
            return bytes.fromhex(value)
        elif type_ == 'date':
            return datetime.date.fromisoformat(value)
        elif type_ == 'dateTime':
            return parse_datetime(value)
        elif type_ == 'duration':
            return parse_duration(value)

        # return to string
        return value

    @handle_namespace(NAMESPACE_PRE_32, False)
    def gml_pre_32(self, node: Element):
        if is_temporal(node):
            return parse_temporal(node, NSMAP_PRE_32)

        return values.Geometry(parse_pre_v32(node))

    @handle_namespace(NAMESPACE_32, False)
    def gml_32(self, node: Element):
        if is_temporal(node):
            return parse_temporal(node, NSMAP_32)

        return values.Geometry(parse_v32(node))

    @handle_namespace(NAMESPACE_33_CE, False)
    def gml_33_ce(self, node: Element):
        return values.Geometry(parse_v33_ce(node))

    @handle_namespace(NAMESPACE_GEORSS, False)
    def georss(self, node: Element):
        return values.Geometry(parse_georss(node))



from operator import (
    and_, or_, lt, le, gt, ge, ne, eq, add, sub, mul, truediv
)
from functools import reduce

import shapely

from ...util import like_pattern_to_re


def combine(sub_filters, combinator: str):
    """ Combine filters using a logical combinator """
    assert combinator in ("AND", "OR")
    op = and_ if combinator == "AND" else or_
    return reduce(
        lambda acc, q: op(acc, q) if acc is not None else q,
        sub_filters
    )


def negate(sub_filter):
    """ Negate a filter, opposing its meaning. """
    return ~sub_filter


OP_MAP = {
    "<": lt,
    "<=": le,
    ">": gt,
    ">=": ge,
    "<>": ne,
    "=": eq,
}


def compare(lhs, rhs, op):
    return OP_MAP[op](lhs, rhs)


def between(lhs, low, high, not_):
    result = lhs.between(low, high)
    if not_:
        result = ~result
    return result


def like(lhs, pattern, nocase, wildcard, singlechar, escapechar, not_):
    regex = like_pattern_to_re(
        pattern,
        nocase,
        wildcard,
        singlechar,
        escapechar or '\\'
    )
    result = lhs.str.match(regex)
    if not_:
        result = ~result
    return result


def contains(lhs, items, not_):
    # TODO: check if dataframe or scalar
    result = lhs.isin(items)
    if not_:
        result = ~result
    return result


def null(lhs, not_):
    result = lhs.isnull()
    if not_:
        result = ~result
    return result


def temporal(lhs, time_or_period, op):
    pass
    # TODO implement


SPATIAL_OP_MAP = {
    "INTERSECTS": 'intersects',
    "DISJOINT": 'disjoint',
    "CONTAINS": 'contains',
    "WITHIN": 'within',
    "TOUCHES": 'touches',
    "CROSSES": 'crosses',
    "OVERLAPS": 'overlaps',
    "EQUALS": 'geom_equals',
}


def spatial(lhs, rhs, op):
    assert op in SPATIAL_OP_MAP
    return getattr(lhs, SPATIAL_OP_MAP[op])(rhs)


def bbox(lhs, minx, miny, maxx, maxy, crs=None):
    box = shapely.geometry.Polygon.from_bounds(minx, miny, maxx, maxy)
    # TODO: handle CRS
    return lhs.intersects(box)


def attribute(df, name, field_mapping=None):
    if field_mapping:
        name = field_mapping[name]
    return df[name]


OP_TO_FUNC = {
    "+": add,
    "-": sub,
    "*": mul,
    "/": truediv
}


def arithmetic(lhs, rhs, op):
    """ Create an arithmetic filter """
    return OP_TO_FUNC[op](lhs, rhs)



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from typing import Any, Dict, Callable, Optional, Tuple, Union

from datetime import date, time, datetime, timedelta, timezone

import shapely.geometry

from ... import ast
from ... import values
from ...util import like_pattern_to_re, parse_datetime

from ..evaluator import Evaluator, handle


COMPARISON_MAP = {
    ast.ComparisonOp.EQ: '==',
    ast.ComparisonOp.NE: '!=',
    ast.ComparisonOp.LT: '<',
    ast.ComparisonOp.LE: '<=',
    ast.ComparisonOp.GT: '>',
    ast.ComparisonOp.GE: '>=',
}

ARITHMETIC_MAP = {
    ast.ArithmeticOp.ADD: '+',
    ast.ArithmeticOp.SUB: '-',
    ast.ArithmeticOp.MUL: '*',
    ast.ArithmeticOp.DIV: '/',
}

ARRAY_COMPARISON_OP_MAP = {
    ast.ArrayComparisonOp.AEQUALS: '==',
    ast.ArrayComparisonOp.ACONTAINS: '>=',
    ast.ArrayComparisonOp.ACONTAINEDBY: '<=',
    ast.ArrayComparisonOp.AOVERLAPS: '&',
}


class NativeEvaluator(Evaluator):
    """ This evaluator type allows to create a filter that can be used to
        filter objects or dicts.

        The filter is built using Python expressions which are then parsed
        using eval. The result is a callable object that can be used in any
        circumstance a normal function would.
        The callable object accepts a single parameter: the object to filter
        and returns a boolean if the object matches the filters or not.
    """

    def __init__(self, function_map: Dict[str, Callable] = None,
                 attribute_map: Dict[str, str] = None,
                 use_getattr: bool = True,
                 allow_nested_attributes: bool = True):
        """ Constructs a NativeEvaluator.

            Args:
                function_map: a mapping of a function name to a callable
                    function.
                attribute_map: a mapping of an external name to an internal
                    field of the item to be filtered. The internal field
                    specifier can be a JSON-Path that will be resolved against
                    the passed in item.
        """
        self.function_map = function_map if function_map is not None else {}
        self.attribute_map = attribute_map
        self.use_getattr = use_getattr
        self.allow_nested_attributes = allow_nested_attributes
        self.locals: Dict[str, Any] = {}
        self.local_count = 0

    def _add_local(self, value: Any) -> str:
        " Add a value as a local variable to the expression. "
        self.local_count += 1
        key = f'local_{self.local_count}'
        self.locals[key] = value
        return key

    def _resolve_attribute(self, name):
        """ Helper to resolve an attribute, either directly or via the
            integrated ``attribute_map``
        """
        if self.attribute_map is not None:
            if name in self.attribute_map:
                path = self.attribute_map[name]
            elif '*' in self.attribute_map:
                path = self.attribute_map['*'].replace('*', name)
            allow_nested_attributes = True
        else:
            path = name
            allow_nested_attributes = self.allow_nested_attributes

        parts = path.split('.')
        if not allow_nested_attributes and len(parts) > 1:
            raise Exception('Nested attributes are not allowed')

        return parts

    @handle(ast.Not)
    def not_(self, node, sub):
        return f'(not {sub})'

    @handle(ast.And)
    def and_(self, node, lhs, rhs):
        return f'({lhs} and {rhs})'

    @handle(ast.Or)
    def or_(self, node, lhs, rhs):
        return f'({lhs} or {rhs})'

    @handle(ast.Comparison, subclasses=True)
    def comparison(self, node, lhs, rhs):
        op = COMPARISON_MAP[node.op]
        return f'({lhs} {op} {rhs})'

    @handle(ast.Between)
    def between(self, node, lhs, low, high):
        if node.not_:
            return f'({low} > {lhs} or {lhs} > {high})'
        else:
            return f'({low} <= {lhs} <= {high})'

    @handle(ast.Like)
    def like(self, node, lhs):
        maybe_not_inv = '' if node.not_ else 'not '
        regex = like_pattern_to_re(
            node.pattern,
            node.nocase,
            node.wildcard,
            node.singlechar,
            node.escapechar
        )
        key = self._add_local(regex)
        return f'({key}.match({lhs}) is {maybe_not_inv}None)'

    @handle(ast.In)
    def in_(self, node, lhs, *options):
        maybe_not = 'not' if node.not_ else ''
        opts = ', '.join([
            f'{opt}' for opt in options
        ])
        return f'({lhs} {maybe_not} in ({opts}))'

    @handle(ast.IsNull)
    def null(self, node, lhs):
        maybe_not = 'not ' if node.not_ else ''
        return f'({lhs} is {maybe_not}None)'

    @handle(ast.Exists)
    def exists(self, node, lhs):
        parts = self._resolve_attribute(node.lhs.name)
        maybe_not = 'not ' if node.not_ else ''
        if self.use_getattr:
            cur = 'item'
            for part in parts[:-1]:
                cur = f'getattr({cur}, {part!r}, None)'
            return f'({maybe_not}hasattr({cur}, {parts[-1]!r}))'
        else:
            getters = ''.join(
                f'.get({part!r}, {{}})'
                for part in parts[:-1]
            )
            return f'{parts[-1]!r} {maybe_not}in item{getters}'

    @handle(ast.TemporalPredicate, subclasses=True)
    def temporal(self, node, lhs, rhs):
        return (
            f'(relate_intervals(to_interval({lhs}),'
            f'to_interval({rhs})) == '
            f'ast.TemporalComparisonOp.{node.op.name})'
        )

    @handle(ast.ArrayPredicate, subclasses=True)
    def array(self, node, lhs, rhs):
        op = ARRAY_COMPARISON_OP_MAP[node.op]
        return f'bool(set({lhs}) {op} set({rhs}))'

    @handle(ast.SpatialComparisonPredicate, subclasses=True)
    def spatial_operation(self, node, lhs, rhs):
        return (
            f'(getattr(ensure_spatial({lhs}), '
            f'{node.op.value.lower()!r})({rhs}))'
        )

    @handle(ast.Relate)
    def spatial_pattern(self, node, lhs, rhs):
        return (
            f'(ensure_spatial({lhs}).relate_pattern({rhs}, {node.pattern!r}))'
        )

    @handle(ast.BBox)
    def bbox(self, node, lhs):
        bbox_local = self._add_local(shapely.geometry.Polygon.from_bounds(
            node.minx, node.miny, node.maxx, node.maxy
        ))
        return f'(ensure_spatial({lhs}).intersects({bbox_local}))'

    @handle(ast.Attribute)
    def attribute(self, node):
        parts = self._resolve_attribute(node.name)
        if self.use_getattr:
            cur = 'item'
            for part in parts:
                cur = f'getattr({cur}, {part!r}, None)'
            return cur
        else:
            getters = ''.join(
                f'.get({part!r})'
                for part in parts
            )
            return f'item{getters}'

    @handle(ast.Arithmetic, subclasses=True)
    def arithmetic(self, node, lhs, rhs):
        op = ARITHMETIC_MAP[node.op]
        return f'({lhs}) {op} ({rhs})'

    @handle(ast.Function)
    def function(self, node, *arguments):
        args = ', '.join([
            f'({arg})' for arg in arguments
        ])
        return f'{node.name}({args})'

    @handle(*values.LITERALS)
    def literal(self, node):
        key = self._add_local(node)
        return key

    @handle(values.Interval)
    def interval(self, node, low, high):
        return f'values.Interval({low}, {high})'

    @handle(values.Geometry)
    def geometry(self, node):
        key = self._add_local(shapely.geometry.shape(node))
        return key

    @handle(values.Envelope)
    def envelope(self, node):
        key = self._add_local(
            shapely.geometry.Polygon.from_bounds(
                node.x1, node.y1, node.x2, node.y2
            )
        )
        return key

    def adopt_result(self, result):
        """ Turns the compiled expression into a callable object using
            ``eval``. Literals are passed in as well as the function map.
        """
        expression = f'lambda item: {result}'
        globals_ = {
            'relate_intervals': relate_intervals,
            'to_interval': to_interval,
            'ensure_spatial': ensure_spatial,
            'ast': ast,
            'values': values,
        }
        if not set(globals_).isdisjoint(set(self.function_map)):
            raise ValueError(
                f"globals collision {list(globals_)} and "
                f"{list(self.function_map)}"
            )

        globals_.update(self.function_map)
        globals_.update(self.locals)

        # clear any locals for later use
        self.locals.clear()

        return eval(expression, globals_)


MaybeInterval = Union[values.Interval, date, datetime, str, None]
InternalInterval = Tuple[Optional[datetime], Optional[datetime]]


def to_interval(value: MaybeInterval) -> InternalInterval:
    """ Converts the given value to an interval tuple of ``start``/``stop``
        as Python datetime objects.

        - ``values.Interval`` objects are expanded to two datetimes:
            - two datetimes are returned as such
            - a date is transformed to a datetime, where the ``time``
              component is either ``time.min`` for start or ``time.max``
              for then end component.
            - if either the start or end is a ``timedelta`` object, that value
              is either added to the start value or subtracted from the end
              value.
        - ``date`` objects are transformed to two datetimes for the
          ``time.min`` and ``time.end`` of that date in UTC.
        - ``datetime`` and ``str`` objects are an interval with both
          start and end of the same value. Strings are parsed beforehand.
        - ``None`` is simply returned as ``(None, None)``
    """

    if isinstance(value, str):
        value = parse_datetime(value)

    if isinstance(value, values.Interval):
        low = value.start
        high = value.end

        # convert low and high dates to their respective datetime
        # by using 00:00 time for the low part and 23:59:59 for the high
        # part
        if isinstance(low, date):
            low = datetime.combine(low, time.min, timezone.utc)
        if isinstance(high, date):
            high = datetime.combine(high, time.max, timezone.utc)

        # low and high are now either datetimes, timedeltas or None

        if isinstance(low, timedelta):
            if isinstance(high, datetime):
                low = high - low
            else:
                raise ValueError(f'Cannot combine {low} with {high}')
        elif isinstance(high, timedelta):
            if isinstance(low, datetime):
                high = low + high
            else:
                raise ValueError(f'Cannot combine {low} with {high}')

        return (low, high)

    elif isinstance(value, date):
        return (
            datetime.combine(value, time.min, timezone.utc),
            datetime.combine(value, time.max, timezone.utc),
        )

    elif isinstance(value, datetime):
        return (value, value)

    elif value is None:
        return (None, None)

    raise ValueError(f'Invalid type {type(value)}')


def relate_intervals(lhs: InternalInterval,
                     rhs: InternalInterval) -> ast.TemporalComparisonOp:
    """ Relates two intervals (tuples of two ``datetime`` or ``None`` values)
        and returns the associated ``ast.TemporalComparisonOp`` value.
    """
    ll, lh = lhs
    rl, rh = rhs
    if ll is None or lh is None or rl is None or rh is None:
        # TODO: handle open ended intervals (None on either side)
        return ast.TemporalComparisonOp.DISJOINT
    elif lh < rl:
        return ast.TemporalComparisonOp.BEFORE
    elif ll > rh:
        return ast.TemporalComparisonOp.AFTER
    elif lh == rl:
        return ast.TemporalComparisonOp.MEETS
    elif ll == rh:
        return ast.TemporalComparisonOp.METBY
    elif ll < rl and rl < lh < rh:
        return ast.TemporalComparisonOp.TOVERLAPS
    elif rl < ll < rh and lh > rh:
        return ast.TemporalComparisonOp.OVERLAPPEDBY
    elif ll == rl and lh < rh:
        return ast.TemporalComparisonOp.BEGINS
    elif ll == rl and lh > rh:
        return ast.TemporalComparisonOp.BEGUNBY
    elif ll > rl and lh < rh:
        return ast.TemporalComparisonOp.DURING
    elif ll < rl and lh > rh:
        return ast.TemporalComparisonOp.TCONTAINS
    elif ll > rl and lh == rh:
        return ast.TemporalComparisonOp.ENDS
    elif ll < rl and lh == rh:
        return ast.TemporalComparisonOp.ENDEDBY
    elif ll == rl and lh == rh:
        return ast.TemporalComparisonOp.TEQUALS

    raise ValueError(
        f'Error relating intervals [{ll}, {lh}] and [{rl}, {rh}]'
    )


def ensure_spatial(value: Any) -> shapely.geometry.base.BaseGeometry:
    """ Ensures that a given value is a shapely geometry. If it is already
        it is passed through, otherwise it is tried to be parsed via
        ``shapely.geometry.shape``.
    """
    if isinstance(value, shapely.geometry.base.BaseGeometry):
        return value
    return shapely.geometry.shape(value)



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>,
# David Bitner <bitner@dbspatial.com>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from datetime import date, datetime, timedelta
from typing import List, Union, cast
import json

from ... import ast
from ... import values
from ...util import parse_datetime, parse_date, parse_duration
from ...cql2 import BINARY_OP_PREDICATES_MAP

# https://github.com/opengeospatial/ogcapi-features/tree/master/cql2


JsonType = Union[dict, list, str, float, int, bool, None]


def walk_cql_json(node: JsonType):
    if isinstance(
        node,
        (
            str,
            float,
            int,
            bool,
            datetime,
            values.Geometry,
            values.Interval,
            ast.Node,
        ),
    ):
        return node

    if isinstance(node, list):
        return [walk_cql_json(sub_node) for sub_node in node]

    if not isinstance(node, dict):
        raise ValueError(f"Invalid type {type(node)}")

    if "filter-lang" in node and node["filter-lang"] != "cql2-json":
        raise Exception(f"Cannot parse {node['filter-lang']} with cql2-json.")

    elif "filter" in node:
        return walk_cql_json(node["filter"])

    # check if we are dealing with a geometry
    if "type" in node and "coordinates" in node:
        # TODO: test if node is actually valid
        return values.Geometry(node)

    elif "bbox" in node:
        return values.Envelope(*node["bbox"])

    elif "date" in node:
        return parse_date(node["date"])

    elif "timestamp" in node:
        return parse_datetime(node["timestamp"])

    elif "interval" in node:
        parsed: List[Union[date, datetime, timedelta, None]] = []
        for value in node["interval"]:
            if value == "..":
                parsed.append(None)
                continue
            try:
                parsed.append(parse_date(value))
            except ValueError:
                try:
                    parsed.append(parse_duration(value))
                except ValueError:
                    parsed.append(parse_datetime(value))

        return values.Interval(*parsed)

    elif "property" in node:
        return ast.Attribute(node["property"])

    elif "function" in node:
        return ast.Function(
            node["function"]["name"],
            cast(
                List[ast.AstType], walk_cql_json(node["function"]["arguments"])
            ),
        )

    elif "lower" in node:
        return ast.Function(
            "lower", [cast(ast.Node, walk_cql_json(node["lower"]))]
        )

    elif "op" in node:
        op = node["op"]
        args = walk_cql_json(node["args"])

        if op in ("and", "or"):
            return (ast.And if op == "and" else ast.Or).from_items(*args)

        elif op == "not":
            # allow both arrays and objects, the standard is ambigous in
            # that regard
            if isinstance(args, list):
                args = args[0]
            return ast.Not(cast(ast.Node, walk_cql_json(args)))

        elif op == "isNull":
            return ast.IsNull(cast(ast.Node, walk_cql_json(args)), False)

        elif op == "between":
            return ast.Between(
                cast(ast.Node, walk_cql_json(args[0])),
                cast(ast.ScalarAstType, walk_cql_json(args[1][0])),
                cast(ast.ScalarAstType, walk_cql_json(args[1][1])),
                not_=False,
            )

        elif op == "like":
            return ast.Like(
                cast(ast.Node, walk_cql_json(args[0])),
                pattern=cast(str, args[1]),
                nocase=False,
                wildcard="%",
                singlechar=".",
                escapechar="\\",
                not_=False,
            )

        elif op == "in":
            return ast.In(
                cast(ast.AstType, walk_cql_json(args[0])),
                cast(List[ast.AstType], walk_cql_json(args[1])),
                not_=False,
            )

        elif op == "isNull":
            return ast.IsNull(
                walk_cql_json(args),
                not_=False,
            )

        elif op in BINARY_OP_PREDICATES_MAP:
            args = [
                cast(ast.Node, walk_cql_json(arg)) for arg in args
            ]
            return BINARY_OP_PREDICATES_MAP[op](
                *args
            )

    raise ValueError(f"Unable to parse expression node {node!r}")


def parse(cql: Union[str, dict]) -> ast.AstType:
    if isinstance(cql, str):
        root = json.loads(cql)
    else:
        root = cql

    return walk_cql_json(root)



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

"""
Parser implementation for JFE. Spec here:
https://github.com/tschaub/ogcapi-features/tree/json-array-expression/extensions/cql/jfe
"""

from datetime import datetime
import json
from typing import Any, Dict, List, Type, Union, cast

from ... import ast
from ... import values
from ...util import parse_datetime


COMPARISON_MAP: Dict[str, Type] = {
    '==': ast.Equal,
    '!=': ast.NotEqual,
    '<': ast.LessThan,
    '<=': ast.LessEqual,
    '>': ast.GreaterThan,
    '>=': ast.GreaterEqual,
}

SPATIAL_PREDICATES_MAP = {
    'intersects': ast.GeometryIntersects,
    'within': ast.GeometryWithin,
}

TEMPORAL_PREDICATES_MAP = {
    'before': ast.TimeBefore,
    'after': ast.TimeAfter,
    'during': ast.TimeDuring,
}

ARITHMETIC_MAP = {
    '+': ast.Add,
    '-': ast.Sub,
    '*': ast.Mul,
    '/': ast.Div,
}

FUNCTION_MAP = {
    '%': 'mod',
    '^': 'pow',
}

ParseResult = Union[
    ast.Node,
    str,
    float,
    int,
    datetime,
    values.Geometry,
    values.Interval,
    Dict[Any, Any]  # TODO: for like wildcards.
]


def _parse_node(node: Union[list, dict]) -> ParseResult:
    if isinstance(node, (str, float, int)):
        return node
    elif isinstance(node, dict):
        # wrap geometry, we say that the 'type' property defines if it is a
        # geometry
        if 'type' in node:
            return values.Geometry(node)

        # just return objects for example `like` wildcards
        else:
            return node

    if not isinstance(node, list):
        raise ValueError(f'Invalid node class {type(node)}')

    op = node[0]
    arguments = [_parse_node(sub) for sub in node[1:]]

    if op in ['all', 'any']:
        cls = ast.And if op == 'all' else ast.Or
        return cls.from_items(*arguments)

    elif op == '!':
        return ast.Not(*cast(List[ast.Node], arguments))

    elif op in COMPARISON_MAP:
        return COMPARISON_MAP[op](*arguments)

    elif op == 'like':
        wildcard = '%'
        if len(arguments) > 2:
            wildcard = cast(dict, arguments[2]).get('wildCard', '%')
        return ast.Like(
            cast(ast.Node, arguments[0]),
            cast(str, arguments[1]),
            nocase=False,
            wildcard=wildcard,
            singlechar='.',
            escapechar='\\',
            not_=False,
        )

    elif op == 'in':
        assert isinstance(arguments[0], ast.Node)
        return ast.In(
            cast(ast.Node, arguments[0]),
            cast(List[ast.AstType], arguments[1:]),
            not_=False
        )

    elif op in SPATIAL_PREDICATES_MAP:
        return SPATIAL_PREDICATES_MAP[op](
            *cast(List[ast.SpatialAstType], arguments)
        )

    elif op in TEMPORAL_PREDICATES_MAP:
        # parse strings to datetimes
        dt_args = [
            parse_datetime(arg) if isinstance(arg, str) else arg
            for arg in arguments
        ]
        if len(arguments) == 3:
            if isinstance(dt_args[0], datetime) and \
                    isinstance(dt_args[1], datetime):
                dt_args = [
                    values.Interval(dt_args[0], dt_args[1]),
                    dt_args[2],
                ]
            if isinstance(dt_args[1], datetime) and \
                    isinstance(dt_args[2], datetime):
                dt_args = [
                    dt_args[0],
                    values.Interval(dt_args[1], dt_args[2]),
                ]

        return TEMPORAL_PREDICATES_MAP[op](
            *cast(List[ast.TemporalAstType], dt_args)
        )

    # special property getters
    elif op in ['id', 'geometry']:
        return ast.Attribute(op)

    # normal property getter
    elif op == 'get':
        return ast.Attribute(arguments[0])

    elif op == 'bbox':
        pass  # TODO

    elif op in ARITHMETIC_MAP:
        return ARITHMETIC_MAP[op](
            *cast(List[ast.ScalarAstType], arguments)
        )

    elif op in ['%', 'floor', 'ceil', 'abs', '^', 'min', 'max']:
        return ast.Function(
            FUNCTION_MAP.get(op, op), cast(List[ast.AstType], arguments)
        )

    raise ValueError(f'Invalid expression operation \'{op}\'')


def parse(jfe: Union[str, list, dict]) -> ast.Node:
    """ Parses the given JFE expression (either a string or an already
        parsed JSON) to an AST.
        If a string is passed, it will be parsed as JSON.

        https://github.com/tschaub/ogcapi-features/tree/json-array-expression/extensions/cql/jfe
    """
    if isinstance(jfe, str):
        root = json.loads(jfe)
    else:
        root = jfe

    return cast(ast.Node, _parse_node(root))



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from typing import List, Union, cast
import json

from ...util import parse_duration, parse_datetime
from ...values import Envelope, Geometry
from ... import ast
from ... import values
from datetime import datetime

# https://portal.ogc.org/files/96288


COMPARISON_MAP = {
    'eq': ast.Equal,
    'lt': ast.LessThan,
    'lte': ast.LessEqual,
    'gt': ast.GreaterThan,
    'gte': ast.GreaterEqual,
}

SPATIAL_PREDICATES_MAP = {
    'intersects': ast.GeometryIntersects,
    'equals': ast.GeometryEquals,
    'disjoint': ast.GeometryDisjoint,
    'touches': ast.GeometryTouches,
    'within': ast.GeometryWithin,
    'overlaps': ast.GeometryOverlaps,
    'crosses': ast.GeometryCrosses,
    'contains': ast.GeometryContains,
}

TEMPORAL_PREDICATES_MAP = {
    'before': ast.TimeBefore,
    'after': ast.TimeAfter,
    'meets': ast.TimeMeets,
    'metby': ast.TimeMetBy,
    'toverlaps': ast.TimeOverlaps,
    'overlappedby': ast.TimeOverlappedBy,
    'begins': ast.TimeBegins,
    'begunby': ast.TimeBegunBy,
    'during': ast.TimeDuring,
    'tcontains': ast.TimeContains,
    'ends': ast.TimeEnds,
    'endedby': ast.TimeEndedBy,
    'tequals': ast.TimeEquals,
    # 'anyinteract': ast.TimeAnyInteract, # TODO?
}


ARRAY_PREDICATES_MAP = {
    'aequals': ast.ArrayEquals,
    'acontains': ast.ArrayContains,
    'acontainedBy': ast.ArrayContainedBy,
    'aoverlaps': ast.ArrayOverlaps,
}

ARITHMETIC_MAP = {
    '+': ast.Add,
    '-': ast.Sub,
    '*': ast.Mul,
    '/': ast.Div,
}


def walk_cql_json(node: dict, is_temporal: bool = False) -> ast.AstType:
    if is_temporal and isinstance(node, str):
        # Open interval
        if node == '..':
            return None

        try:
            return parse_duration(node)
        except ValueError:
            value = parse_datetime(node)

        if value is None:
            raise ValueError(f'Failed to parse temporal value from {node}')

        return value

    if isinstance(node, (str, float, int, bool)):
        return node

    if isinstance(node, list):
        result = [
            cast(datetime, walk_cql_json(sub_node, is_temporal))
            for sub_node in node
        ]
        if is_temporal:
            return values.Interval(*result)
        else:
            return result

    assert isinstance(node, dict)

    # check if we are dealing with a geometry
    if 'type' in node and 'coordinates' in node:
        # TODO: test if node is actually valid
        return Geometry(node)

    elif 'bbox' in node:
        return Envelope(*node['bbox'])

    # decode all other nodes
    for name, value in node.items():
        if name in ('and', 'or'):
            sub_items = cast(list, walk_cql_json(value))
            return (ast.And if name == 'and' else ast.Or).from_items(
                *sub_items
            )

        elif name == 'not':
            # allow both arrays and objects, the standard is ambigous in
            # that regard
            if isinstance(value, list):
                value = value[0]
            return ast.Not(cast(ast.Node, walk_cql_json(value)))

        elif name in COMPARISON_MAP:
            return COMPARISON_MAP[name](
                cast(ast.ScalarAstType, walk_cql_json(value[0])),
                cast(ast.ScalarAstType, walk_cql_json(value[1])),
            )

        elif name == 'between':
            return ast.Between(
                cast(ast.Node, walk_cql_json(value['value'])),
                cast(ast.ScalarAstType, walk_cql_json(value['lower'])),
                cast(ast.ScalarAstType, walk_cql_json(value['upper'])),
                not_=False,
            )

        elif name == 'like':
            return ast.Like(
                cast(ast.Node, walk_cql_json(value['like'][0])),
                cast(str, value['like'][1]),
                nocase=value.get('nocase', True),
                wildcard=value.get('wildcard', '%'),
                singlechar=value.get('singleChar', '.'),
                escapechar=value.get('escapeChar', '\\'),
                not_=False,
            )

        elif name == 'in':
            return ast.In(
                cast(ast.AstType, walk_cql_json(value['value'])),
                cast(List[ast.AstType], walk_cql_json(value['list'])),
                not_=False,
                # TODO nocase
            )

        elif name == 'isNull':
            return ast.IsNull(
                walk_cql_json(value),
                not_=False,
            )

        elif name in SPATIAL_PREDICATES_MAP:
            return SPATIAL_PREDICATES_MAP[name](
                cast(ast.SpatialAstType, walk_cql_json(value[0])),
                cast(ast.SpatialAstType, walk_cql_json(value[1])),
            )

        elif name in TEMPORAL_PREDICATES_MAP:
            return TEMPORAL_PREDICATES_MAP[name](
                cast(
                    ast.TemporalAstType,
                    walk_cql_json(value[0], is_temporal=True)
                ),
                cast(
                    ast.TemporalAstType,
                    walk_cql_json(value[1], is_temporal=True)
                ),
            )

        elif name in ARRAY_PREDICATES_MAP:
            return ARRAY_PREDICATES_MAP[name](
                cast(ast.ArrayAstType, walk_cql_json(value[0])),
                cast(ast.ArrayAstType, walk_cql_json(value[1])),
            )

        elif name in ARITHMETIC_MAP:
            return ARITHMETIC_MAP[name](
                cast(ast.ScalarAstType, walk_cql_json(value[0])),
                cast(ast.ScalarAstType, walk_cql_json(value[1])),
            )

        elif name == 'property':
            return ast.Attribute(value)

        elif name == 'function':
            return ast.Function(
                value['name'],
                cast(List[ast.AstType], walk_cql_json(value['arguments'])),
            )

    raise ValueError(f'Unable to parse expression node {node!r}')


def parse(cql: Union[str, dict]) -> ast.AstType:
    if isinstance(cql, str):
        root = json.loads(cql)
    else:
        root = cql

    return walk_cql_json(root)



import re
from datetime import timedelta
from functools import reduce
from inspect import signature
from typing import Callable, Dict

from sqlalchemy import and_, func, not_, or_
from pygeoif.geometry import as_shape


def parse_bbox(box, srid: int = None):
    minx, miny, maxx, maxy = box
    return func.ST_GeomFromEWKT(
        f"SRID={4326 if srid is None else srid};POLYGON(("
        f"{minx} {miny}, {minx} {maxy}, "
        f"{maxx} {maxy}, {maxx} {miny}, "
        f"{minx} {miny}))"
    )


def parse_geometry(geom):
    wkt = as_shape(geom).to_wkt()
    search = re.search(r"SRID=(\d+);", wkt)
    sridtxt = "" if search else "SRID=4326;"
    return func.ST_GeomFromEWKT(f"{sridtxt}{wkt}")


# ------------------------------------------------------------------------------
# Filters
# ------------------------------------------------------------------------------
class Operator:

    OPERATORS: Dict[str, Callable] = {
        "is_null": lambda f, a=None: f.is_(None),
        "is_not_null": lambda f, a=None: f.isnot(None),
        "==": lambda f, a: f == a,
        "=": lambda f, a: f == a,
        "eq": lambda f, a: f == a,
        "!=": lambda f, a: f != a,
        "<>": lambda f, a: f != a,
        "ne": lambda f, a: f != a,
        ">": lambda f, a: f > a,
        "gt": lambda f, a: f > a,
        "<": lambda f, a: f < a,
        "lt": lambda f, a: f < a,
        ">=": lambda f, a: f >= a,
        "ge": lambda f, a: f >= a,
        "<=": lambda f, a: f <= a,
        "le": lambda f, a: f <= a,
        "like": lambda f, a: f.like(a),
        "ilike": lambda f, a: f.ilike(a),
        "not_ilike": lambda f, a: ~f.ilike(a),
        "in": lambda f, a: f.in_(a),
        "not_in": lambda f, a: ~f.in_(a),
        "any": lambda f, a: f.any(a),
        "not_any": lambda f, a: func.not_(f.any(a)),
        "INTERSECTS": lambda f, a: f.ST_Intersects(a),
        "DISJOINT": lambda f, a: f.ST_Disjoint(a),
        "CONTAINS": lambda f, a: f.ST_Contains(a),
        "WITHIN": lambda f, a: f.ST_Within(a),
        "TOUCHES": lambda f, a: f.ST_Touches(a),
        "CROSSES": lambda f, a: f.ST_Crosses(a),
        "OVERLAPS": lambda f, a: f.ST_Overlaps(a),
        "EQUALS": lambda f, a: f.ST_Equals(a),
        "RELATE": lambda f, a, pattern: f.ST_Relate(a, pattern),
        "DWITHIN": lambda f, a, distance: f.ST_Dwithin(a, distance),
        "BEYOND": lambda f, a, distance: ~f.ST_Dwithin(a, distance),
        "+": lambda f, a: f + a,
        "-": lambda f, a: f - a,
        "*": lambda f, a: f * a,
        "/": lambda f, a: f / a,
    }

    def __init__(self, operator: str = None):
        if not operator:
            operator = "=="

        if operator not in self.OPERATORS:
            raise Exception("Operator `{}` not valid.".format(operator))

        self.operator = operator
        self.function = self.OPERATORS[operator]
        self.arity = len(signature(self.function).parameters)


def combine(sub_filters, combinator: str = "AND"):
    """ Combine filters using a logical combinator

        :param sub_filters: the filters to combine
        :param combinator: a string: "AND" / "OR"
        :return: the combined filter
    """
    assert combinator in ("AND", "OR")
    _op = and_ if combinator == "AND" else or_

    def test(acc, q):
        return _op(acc, q)

    return reduce(test, sub_filters)


def negate(sub_filter):
    """ Negate a filter, opposing its meaning.

        :param sub_filter: the filter to negate
        :return: the negated filter
    """
    return not_(sub_filter)


def runop(lhs, rhs=None, op: str = "=", negate: bool = False):
    """ Compare a filter with an expression using a comparison operation

        :param lhs: the field to compare
        :param rhs: the filter expression
        :param op: a string denoting the operation.
        :return: a comparison expression object
    """
    _op = Operator(op)

    if negate:
        return not_(_op.function(lhs, rhs))
    return _op.function(lhs, rhs)


def between(lhs, low, high, negate=False):
    """ Create a filter to match elements that have a value within a certain
        range.

        :param lhs: the field to compare
        :param low: the lower value of the range
        :param high: the upper value of the range
        :param not_: whether the range shall be inclusive (the default) or
                     exclusive
        :return: a comparison expression object
    """
    l_op = Operator("<=")
    g_op = Operator(">=")
    if negate:
        return not_(and_(g_op.function(lhs, low), l_op.function(lhs, high)))
    return and_(g_op.function(lhs, low), l_op.function(lhs, high))


def like(lhs, rhs, case=False, negate=False):
    """ Create a filter to filter elements according to a string attribute using
        wildcard expressions.

        :param lhs: the field to compare
        :param rhs: the wildcard pattern: a string containing any number of '%'
                    characters as wildcards.
        :param case: whether the lookup shall be done case sensitively or not
        :param not_: whether the range shall be inclusive (the default) or
                     exclusive
        :return: a comparison expression object
    """
    if case:
        _op = Operator("like")
    else:
        _op = Operator("ilike")

    if negate:
        return not_(_op.function(lhs, rhs))
    return _op.function(lhs, rhs)


def temporal(lhs, time_or_period, op):
    """ Create a temporal filter for the given temporal attribute.

        :param lhs: the field to compare
        :type lhs: :class:`django.db.models.F`
        :param time_or_period: the time instant or time span to use as a filter
        :type time_or_period: :class:`datetime.datetime` or a tuple of two
                              datetimes or a tuple of one datetime and one
                              :class:`datetime.timedelta`
        :param op: the comparison operation. one of ``"BEFORE"``,
                   ``"BEFORE OR DURING"``, ``"DURING"``, ``"DURING OR AFTER"``,
                   ``"AFTER"``.
        :type op: str
        :return: a comparison expression object
        :rtype: :class:`django.db.models.Q`
    """
    low = None
    high = None
    equal = None
    if op in ("BEFORE", "AFTER"):
        if op == "BEFORE":
            high = time_or_period
        else:
            low = time_or_period
    elif op == "TEQUALS":
        equal = time_or_period
    else:
        low, high = time_or_period

        if isinstance(low, timedelta):
            low = high - low
        if isinstance(high, timedelta):
            high = low + high
    if low is not None or high is not None:
        if low is not None and high is not None:
            return between(lhs, low, high)
        elif low is not None:
            return runop(lhs, low, ">=")
        else:
            return runop(lhs, high, "<=")
    elif equal is not None:
        return runop(lhs, equal, "==")


UNITS_LOOKUP = {"kilometers": "km", "meters": "m"}


def spatial(lhs, rhs, op, pattern=None, distance=None, units=None):
    """ Create a spatial filter for the given spatial attribute.

        :param lhs: the field to compare
        :param rhs: the time instant or time span to use as a filter
        :param op: the comparison operation. one of ``"INTERSECTS"``,
                   ``"DISJOINT"``, `"CONTAINS"``, ``"WITHIN"``,
                   ``"TOUCHES"``, ``"CROSSES"``, ``"OVERLAPS"``,
                   ``"EQUALS"``, ``"RELATE"``, ``"DWITHIN"``, ``"BEYOND"``
        :param pattern: the spatial relation pattern
        :param distance: the distance value for distance based lookups:
                         ``"DWITHIN"`` and ``"BEYOND"``
        :param units: the units the distance is expressed in
        :return: a comparison expression object
    """

    _op = Operator(op)
    if op == "RELATE":
        return _op.function(lhs, rhs, pattern)
    elif op in ("DWITHIN", "BEYOND"):
        if units == "kilometers":
            distance = distance / 1000
        elif units == "miles":
            distance = distance / 1609
        return _op.function(lhs, rhs, distance)
    else:
        return _op.function(lhs, rhs)


def bbox(lhs, minx, miny, maxx, maxy, crs=4326):
    """ Create a bounding box filter for the given spatial attribute.

        :param lhs: the field to compare
        :param minx: the lower x part of the bbox
        :param miny: the lower y part of the bbox
        :param maxx: the upper x part of the bbox
        :param maxy: the upper y part of the bbox
        :param crs: the CRS the bbox is expressed in
        :return: a comparison expression object
    """

    return lhs.ST_Intersects(parse_bbox([minx, miny, maxx, maxy], crs))


def attribute(name, field_mapping=None):
    """ Create an attribute lookup expression using a field mapping dictionary.

        :param name: the field filter name
        :param field_mapping: the dictionary to use as a lookup.
    """
    field = field_mapping.get(name, name)

    return field


def literal(value):
    return value



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2019 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

"""Install pygeofilter."""

from setuptools import find_packages, setup
import os
import os.path

# don't install dependencies when building win readthedocs
on_rtd = os.environ.get('READTHEDOCS') == 'True'

# get version number
# from https://github.com/mapbox/rasterio/blob/master/setup.py#L55
with open(os.path.join(os.path.dirname(__file__), 'pygeofilter/__init__.py')) as f:
    for line in f:
        if line.find("__version__") >= 0:
            version = line.split("=")[1].strip()
            version = version.strip('"')
            version = version.strip("'")
            break

# use README.md for project long_description
with open('README.md') as f:
    readme = f.read()

setup(
    name='pygeofilter',
    version=version,
    description='pygeofilter is a pure Python parser implementation of OGC filtering standards',
    long_description=readme,
    long_description_content_type="text/markdown",
    author='Fabian Schindler',
    author_email='fabian.schindler@eox.at',
    url='https://github.com/geopython/pygeofilter',
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "dateparser",
        "lark<1.0",
        "pygeoif",
        "dataclasses;python_version<'3.7'",
    ] if not on_rtd else [],
    extras_require={
        'backend-django': ['django'],
        'backend-sqlalchemy': ['geoalchemy2', 'sqlalchemy'],
        'backend-native': ['shapely'],
        'fes': ['pygml>=0.2']
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: GIS',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    tests_require=['pytest']
)



# ------------------------------------------------------------------------------
#
# Project: pygeofilter <https://github.com/geopython/pygeofilter>
# Authors: Fabian Schindler <fabian.schindler@eox.at>,
# David Bitner <bitner@dbspatial.com>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2021 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os.path
import logging

from lark import Lark, logger, v_args

from ... import ast
from ... import values
from ..wkt import WKTTransformer
from ..iso8601 import ISO8601Transformer
from ...cql2 import SPATIAL_PREDICATES_MAP, TEMPORAL_PREDICATES_MAP

logger.setLevel(logging.DEBUG)


@v_args(inline=True)
class CQLTransformer(WKTTransformer, ISO8601Transformer):
    def and_(self, *args):
        return ast.And.from_items(*args)

    def or_(self, *args):
        return ast.Or.from_items(*args)

    def not_(self, node):
        return ast.Not(node)

    def eq(self, lhs, rhs):
        return ast.Equal(lhs, rhs)

    def ne(self, lhs, rhs):
        return ast.NotEqual(lhs, rhs)

    def lt(self, lhs, rhs):
        return ast.LessThan(lhs, rhs)

    def lte(self, lhs, rhs):
        return ast.LessEqual(lhs, rhs)

    def gt(self, lhs, rhs):
        return ast.GreaterThan(lhs, rhs)

    def gte(self, lhs, rhs):
        return ast.GreaterEqual(lhs, rhs)

    def between(self, lhs, low, high):
        return ast.Between(lhs, low, high, False)

    def not_between(self, lhs, low, high):
        return ast.Between(lhs, low, high, True)

    def like(self, node, pattern):
        return ast.Like(node, pattern, False, "%", ".", "\\", False)

    def not_like(self, node, pattern):
        return ast.Like(node, pattern, False, "%", ".", "\\", True)

    def ilike(self, node, pattern):
        return ast.Like(node, pattern, True, "%", ".", "\\", False)

    def not_ilike(self, node, pattern):
        return ast.Like(node, pattern, True, "%", ".", "\\", True)

    def in_(self, node, *options):
        return ast.In(node, list(options), False)

    def not_in(self, node, *options):
        return ast.In(node, list(options), True)

    def null(self, node):
        return ast.IsNull(node, False)

    def not_null(self, node):
        return ast.IsNull(node, True)

    def exists(self, attribute):
        return ast.Exists(attribute, False)

    def does_not_exist(self, attribute):
        return ast.Exists(attribute, True)

    def include(self):
        return ast.Include(False)

    def exclude(self):
        return ast.Include(True)

    def before(self, node, dt):
        return ast.TimeBefore(node, dt)

    def before_or_during(self, node, period):
        return ast.TimeBeforeOrDuring(node, period)

    def during(self, node, period):
        return ast.TimeDuring(node, period)

    def during_or_after(self, node, period):
        return ast.TimeDuringOrAfter(node, period)

    def after(self, node, dt):
        return ast.TimeAfter(node, dt)

    def binary_spatial_predicate(self, op, lhs, rhs):
        op = op.lower()
        return SPATIAL_PREDICATES_MAP[op](lhs, rhs)

    def binary_temporal_predicate(self, lhs, op, rhs):
        op = op.lower()
        return TEMPORAL_PREDICATES_MAP[op](lhs, rhs)

    def relate_spatial_predicate(self, lhs, rhs, pattern):
        return ast.Relate(lhs, rhs, pattern)

    def distance_spatial_predicate(self, op, lhs, rhs, distance, units):
        cls = ast.DistanceWithin if op == "DWITHIN" else ast.DistanceBeyond
        return cls(lhs, rhs, distance, units)

    def distance_units(self, value):
        return value

    def bbox_spatial_predicate(self, lhs, minx, miny, maxx, maxy, crs=None):
        return ast.BBox(lhs, minx, miny, maxx, maxy, crs)

    def function(self, func_name, *expressions):
        name = func_name.name.lower()
        if name == "casei":
            name = "lower"
        return ast.Function(name, list(expressions))

    def add(self, lhs, rhs):
        return ast.Add(lhs, rhs)

    def sub(self, lhs, rhs):
        return ast.Sub(lhs, rhs)

    def mul(self, lhs, rhs):
        return ast.Mul(lhs, rhs)

    def div(self, lhs, rhs):
        return ast.Div(lhs, rhs)

    def neg(self, value):
        return -value

    def attribute(self, name):
        return ast.Attribute(str(name))

    def period(self, start, end):
        return [start, end]

    def INT(self, value):
        return int(value)

    def FLOAT(self, value):
        return float(value)

    def boolean(self, value):
        return value in ("TRUE", "true", "T", "t", "1")

    def DOUBLE_QUOTED(self, token):
        return token[1:-1]

    def SINGLE_QUOTED(self, token):
        return token[1:-1]

    def geometry(self, value):
        return values.Geometry(value)

    def envelope(self, x1, x2, y1, y2):
        return values.Envelope(x1, x2, y1, y2)

    def interval(self, start, end):
        return values.Interval(start, end)


parser = Lark.open(
    "grammar.lark",
    rel_to=__file__,
    parser="lalr",
    debug=True,
    transformer=CQLTransformer(),
    import_paths=[os.path.dirname(os.path.dirname(__file__))],
)


def parse(cql_text):
    return parser.parse(cql_text)


if __name__ == "__main__":
    print(parse("'abc' < 'bce'"))
